from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union, cast

from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.utils.id import pack_id
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.fields import (
    ColumnAlias,
    PrimitiveType,
    RelationAlias,
    is_column,
    is_many_to_many,
    is_one_to_many,
)
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from marshmallow.exceptions import MarshmallowError
from marshmallow.schema import SchemaMeta
from marshmallow_jsonapi import Schema, fields  # type: ignore

CollectionAlias = Union[Collection, CustomizedCollection]


class IntOrFloat(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if value is None:
            return value
        try:
            return int(value)
        except ValueError:
            return float(value)

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        if value is None:
            return value
        try:
            return int(value)
        except ValueError:
            return float(value)


def schema_name(collection: CollectionAlias):
    return f"{collection.name.lower()}_schema"


class JsonApiException(AgentToolkitException):
    pass


def _map_primitive_type(_type: PrimitiveType):
    TYPES = {
        PrimitiveType.STRING: fields.Str,
        PrimitiveType.NUMBER: IntOrFloat,
        PrimitiveType.BOOLEAN: fields.Boolean,
        PrimitiveType.DATE_ONLY: fields.Date,
        PrimitiveType.DATE: fields.DateTime,
        PrimitiveType.TIME_ONLY: fields.Time,
        PrimitiveType.JSON: fields.Raw,
    }
    return TYPES.get(_type, fields.Str)


def _map_attribute_to_marshmallow(column_alias: ColumnAlias):
    if isinstance(column_alias, PrimitiveType):
        return _map_primitive_type(column_alias)()
    elif isinstance(column_alias, list):
        return fields.List
    else:
        return fields.Raw


def _create_relationship(collection: CollectionAlias, field_name: str, relation: RelationAlias):
    many = is_one_to_many(relation) or is_many_to_many(relation)
    if is_many_to_many(relation):
        type_ = relation["through_collection"].lower()
    else:
        type_ = relation["foreign_collection"].lower()

    return ForestRelationShip(
        type_=type_,
        many=many,
        schema=f"{type_}_schema",
        related_url=f"/forest/{collection.name}/{{{collection.name.lower()}_id}}/relationships/{field_name}",
        related_url_kwargs={f"{collection.name.lower()}_id": "<__forest_id__>"},
        collection=collection,
    )


def _create_schema_attributes(collection: CollectionAlias) -> Dict[str, Any]:
    attributes: Dict[str, Any] = {}

    for name, field_schema in collection.schema["fields"].items():
        if is_column(field_schema):
            attributes[name] = _map_attribute_to_marshmallow(field_schema["column_type"])
        else:
            attributes[name] = _create_relationship(collection, name, cast(RelationAlias, field_schema))
    if "id" not in attributes:
        attributes["id"] = fields.Str()
    return attributes


class JsonApiSerializer(type):

    schema: Dict[str, Type["ForestSchema"]] = {}
    attributes: Dict[str, Any] = {}

    def __new__(cls: Any, collection_name: str, bases: Tuple[Any], attrs: Dict[str, Any]):
        cls.attributes[collection_name] = attrs.copy()  # attrs is removed by the parent init
        klass = super(JsonApiSerializer, cls).__new__(cls, collection_name, bases, attrs)
        cls.schema[collection_name] = klass
        return klass

    @classmethod
    def get(cls, collection: CollectionAlias) -> Type["ForestSchema"]:
        json_schema_name = schema_name(collection)
        try:
            json_schema = cls.schema[json_schema_name]
        except KeyError:
            raise JsonApiException(f"The serializer for the collection {collection.name} is not built")

        current_attr = _create_schema_attributes(collection)

        if cls.attributes[json_schema_name].keys() != current_attr.keys():
            json_schema = refresh_json_api_schema(collection)
        return json_schema  # type: ignore


class JsonApiSchemaType(JsonApiSerializer, SchemaMeta):
    pass


class ForestRelationShip(fields.Relationship):
    def __init__(self, *args, **kwargs):  # type: ignore
        self.collection: Collection = kwargs.pop("collection")  # type: ignore
        self.related_collection: Collection = self.collection.datasource.get_collection(kwargs["type_"])  # type: ignore
        super(ForestRelationShip, self).__init__(*args, **kwargs)  # type: ignore

    @property
    def schema(self) -> "ForestSchema":
        SchemaClass: Type[Schema] = JsonApiSerializer.get(self.related_collection)
        return SchemaClass(
            only=getattr(self, "only", None), exclude=getattr(self, "exclude", ()), context=getattr(self, "context", {})
        )

    def get_related_url(self, obj: Any):
        if "data" in obj:
            obj["__forest_id__"] = obj["data"]["id"]
        else:
            obj["__forest_id__"] = obj["id"]
        res: Any = super(ForestRelationShip, self).get_related_url(obj)  # type: ignore
        del obj["__forest_id__"]
        return res


class ForestSchema(Schema):
    def __init__(self, *args, **kwargs):  # type: ignore
        if "projections" in kwargs:
            only, include_data = self._build_only_included_data(kwargs.pop("projections"))  # type: ignore
            kwargs["only"] = only
            if "include_data" not in kwargs:
                kwargs["include_data"] = include_data
        super(ForestSchema, self).__init__(*args, **kwargs)  # type: ignore

    def _build_only_included_data(self, projections: Projection):
        only: Set[str] = set()
        include_data: Set[str] = set()
        for projection in cast(List[str], projections):
            if ":" in projection:
                only.add(projection.replace(":", "."))
                include_data.add(projection.split(":")[0])
            else:
                only.add(projection)
        if "id" not in only:
            only.add("id")
        return only, include_data

    def get_resource_links(self, item: Any):
        item["__forest_id__"] = pack_id(self.Meta.fcollection.schema, item)  # type: ignore
        res = super(ForestSchema, self).get_resource_links(item)  # type: ignore
        del item["__forest_id__"]
        return res

    def _populate_id(self, obj: Union[Dict[str, Any], List[Dict[str, Any]]]):
        if isinstance(obj, list):
            for o in obj:
                self._populate_id(o)
        elif "id" not in obj:
            obj["id"] = pack_id(self.Meta.fcollection.schema, obj)  # type: ignore
        return obj

    def dump(self, obj: Any, *, many: Optional[bool] = None) -> Any:
        self._populate_id(obj)
        try:
            res = super().dump(obj, many=many)  # type: ignore
        except MarshmallowError as e:
            raise JsonApiException(str(e))
        return res  # type: ignore

    def load(self, data, *, many=None, partial=None, unknown=None):  # type: ignore

        try:
            return super().load(data, many=many, partial=partial, unknown=unknown)  # type: ignore
        except MarshmallowError as e:
            raise JsonApiException(str(e))

    def unwrap_item(self, item):  # type: ignore
        # needed to avoid an issue introduced by the front (type are pluralize for add and update)
        item["type"] = self.opts.type_  # type: ignore
        for name, relationship in item.get("relationships", {}).items():  # type: ignore
            relation_field = self.Meta.fcollection.get_field(name)  # type: ignore
            relationship["data"]["type"] = relation_field["foreign_collection"]  # type: ignore
            item["relationships"][name] = relationship
        return super(ForestSchema, self).unwrap_item(item)  # type: ignore


def refresh_json_api_schema(collection: CollectionAlias, ignores: Optional[List[CollectionAlias]] = None):

    if ignores is None:
        ignores = []
    if collection in ignores:
        return
    ignores.append(collection)
    if schema_name(collection) not in JsonApiSerializer.schema:
        raise JsonApiException("The schema doesn't exist")
    del JsonApiSerializer.schema[schema_name(collection)]
    return create_json_api_schema(collection)


def create_json_api_schema(collection: CollectionAlias):
    if schema_name(collection) in JsonApiSerializer.schema:
        raise JsonApiException("The schema has already been created for this collection")

    attributes: Dict[str, Any] = _create_schema_attributes(collection)

    class JsonApiSchema(ForestSchema):
        class Meta:  # type: ignore
            type_ = collection.name
            self_url = f"/forest/{collection.name}/{{{collection.name.lower()}_id}}"
            self_url_kwargs = {f"{collection.name.lower()}_id": "<__forest_id__>"}
            strict = True
            fcollection: CollectionAlias = collection

    return JsonApiSchemaType(schema_name(collection), (JsonApiSchema,), attributes)



from typing import TypeVar, Union

from forestadmin.agent_toolkit.options import Options
from forestadmin.agent_toolkit.resources.base import BaseResource
from forestadmin.agent_toolkit.services.permissions import PermissionService
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.decorators.datasource import CustomizedDatasource

BoundCollection = TypeVar("BoundCollection", bound=Collection)

DatasourceAlias = Union[Datasource[BoundCollection], CustomizedDatasource]


class BaseCollectionResource(BaseResource):
    def __init__(self, datasource: DatasourceAlias[BoundCollection], permission: PermissionService, options: Options):
        super(BaseCollectionResource, self).__init__(options)
        self.permission = permission
        self.datasource = datasource



from typing import Any, Dict, List, Literal, Union

from forestadmin.agent_toolkit.options import AgentMeta, Options
from forestadmin.agent_toolkit.utils.forest_schema.generator_collection import SchemaCollectionGenerator
from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerCollection
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.decorators.datasource import CustomizedDatasource


class SchemaEmitter:
    @staticmethod
    def get_serialized_collection_relation(
        collection: ForestServerCollection, rel_type: Literal["actions", "segments"]
    ):
        data: List[Dict[str, str]] = []
        included: List[Dict[str, str]] = []
        for rel in collection.get(rel_type, []):  # type: ignore
            try:
                name: str = rel["name"]
            except TypeError:
                name: str = rel
            id = f"{collection['name']}.{name}"
            data.append({"id": id, "type": rel_type})
            included.append({"id": id, "type": rel_type, "attributes": rel})

        return data, included

    @classmethod
    def serialize(cls, collections: List[ForestServerCollection], meta: AgentMeta) -> Dict[str, Any]:
        serialized_collections: List[Dict[str, Any]] = []
        included: List[Dict[str, str]] = []
        for collection in collections:
            action_data, action_included = cls.get_serialized_collection_relation(collection, "actions")
            segment_data, segment_included = cls.get_serialized_collection_relation(collection, "segments")
            included.extend(action_included)
            included.extend(segment_included)
            serialized_collections.append(
                {
                    "id": collection["name"],
                    "type": "collections",
                    "attributes": collection,
                    "relationships": {"actions": {"data": action_data}, "segments": {"data": segment_data}},
                }
            )
        return {"data": serialized_collections, "included": included, "meta": meta}

    @classmethod
    async def get_serialized_schema(
        cls, options: Options, datasource: Union[Datasource[Collection], CustomizedDatasource], meta: AgentMeta
    ):
        schema: List[ForestServerCollection] = []
        if not options["is_production"]:
            schema = await SchemaEmitter.generate(options["prefix"], datasource)
        return schema

    @staticmethod
    async def generate(
        prefix: str, datasource: Union[Datasource[Collection], CustomizedDatasource]
    ) -> List[ForestServerCollection]:
        collection_schema: List[ForestServerCollection] = []
        for collection in datasource.collections:
            collection_schema.append(await SchemaCollectionGenerator.build(prefix, collection))
        return collection_schema



from typing import Any, Dict, Optional, TypedDict, TypeVar, Union

from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.resources.collections.exceptions import CollectionResourceException
from forestadmin.agent_toolkit.utils.context import Request, RequestMethod, User
from forestadmin.datasource_toolkit.collections import Collection, CollectionException
from forestadmin.datasource_toolkit.datasources import Datasource, DatasourceException
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.decorators.datasource import CustomizedDatasource
from forestadmin.datasource_toolkit.interfaces.fields import (
    ManyToMany,
    ManyToOne,
    OneToMany,
    OneToOne,
    is_many_to_many,
    is_many_to_one,
    is_one_to_many,
    is_one_to_one,
)
from typing_extensions import Self

BoundCollection = TypeVar("BoundCollection", bound=Collection)


class RequestCollectionException(AgentToolkitException):
    pass


class RequestArgs(TypedDict):
    method: RequestMethod
    collection: Union[Collection, CustomizedCollection]
    body: Optional[Dict[str, Any]]
    query: Optional[Dict[str, str]]
    headers: Optional[Dict[str, str]]
    user: Optional[User]


class RequestCollection(Request):
    def __init__(
        self,
        method: RequestMethod,
        collection: Union[Collection, CustomizedCollection],
        body: Optional[Dict[str, Any]] = None,
        query: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        user: Optional[User] = None,
    ):
        super(RequestCollection, self).__init__(method, body, query, headers, user)
        self.collection = collection

    @staticmethod
    def from_request_args(
        request: Request, datasource: Union[Datasource[BoundCollection], CustomizedDatasource]
    ) -> RequestArgs:
        if not request.query:
            raise RequestCollectionException("'collection_name' is missing in the request")
        try:
            collection_name = request.query["collection_name"]
        except KeyError:
            raise RequestCollectionException("'collection_name' is missing in the request")
        try:
            collection = datasource.get_collection(collection_name)
        except DatasourceException:
            raise RequestCollectionException(f"Collection '{collection_name}' not found")

        return RequestArgs(
            method=request.method,
            body=request.body,
            query=request.query,
            headers=request.headers,
            user=request.user,
            collection=collection,
        )

    @classmethod
    def from_request(
        cls, request: Request, datasource: Union[Datasource[BoundCollection], CustomizedDatasource]
    ) -> Self:
        return cls(**cls.from_request_args(request, datasource))

    @property
    def pks(self):
        if not self.query:
            raise CollectionResourceException("")
        try:
            pks = self.query["pks"]
        except KeyError:
            raise CollectionResourceException("primary keys are missing")
        return pks


class RequestRelationCollection(RequestCollection):
    def __init__(
        self,
        method: RequestMethod,
        collection: Union[Collection, CustomizedCollection],
        foreign_collection: Union[Collection, CustomizedCollection],
        relation: Union[ManyToMany, OneToMany, OneToOne, ManyToOne],
        relation_name: str,
        body: Optional[Dict[str, Any]] = None,
        query: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        user: Optional[User] = None,
    ):
        super(RequestRelationCollection, self).__init__(method, collection, body, query, headers, user)
        self.foreign_collection = foreign_collection
        self.relation = relation
        self.relation_name = relation_name

    @classmethod
    def from_request(
        cls, request: Request, datasource: Union[Datasource[BoundCollection], CustomizedDatasource]
    ) -> Self:
        kwargs = cls.from_request_args(request, datasource)
        if not request.query:
            raise RequestCollectionException("'relation_name' is missing in the request")
        try:
            relation_name = request.query["relation_name"]
        except KeyError:
            raise RequestCollectionException("'relation_name' is missing in the request")

        try:
            related_field = kwargs["collection"].get_field(relation_name)
        except CollectionException:
            raise RequestCollectionException(
                f"{relation_name} is an unknown relation for the collection {kwargs['collection'].name}"
            )

        if (
            is_many_to_one(related_field)
            or is_one_to_one(related_field)
            or is_many_to_many(related_field)
            or is_one_to_many(related_field)
        ):
            try:
                foreign_collection = datasource.get_collection(related_field["foreign_collection"])
            except DatasourceException:
                raise RequestCollectionException(f"Collection '{relation_name}' not found")
            return cls(
                relation=related_field, relation_name=relation_name, foreign_collection=foreign_collection, **kwargs
            )
        else:
            raise RequestCollectionException(f"'{relation_name}' is not a related field")



import enum
import json
from dataclasses import dataclass, field
from io import BytesIO
from typing import Any, Dict, List, Optional


class RequestMethod(enum.Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    OPTION = "OPTION"


@dataclass
class User:
    rendering_id: int
    user_id: int
    tags: Dict[str, Any]
    email: str
    first_name: str
    last_name: str
    team: str


class Request:
    method: RequestMethod
    body: Optional[Dict[str, Any]] = None
    query: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, str]] = None
    user: Optional[User] = None

    def __init__(
        self,
        method: RequestMethod,
        body: Optional[Dict[str, Any]] = None,
        query: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        user: Optional[User] = None,
    ):
        self.method = method
        self.body = body
        self.query = query
        self.headers = headers
        self.user = user


@dataclass
class Response:
    status: int
    body: Optional[str] = None
    headers: Dict[str, str] = field(default_factory=lambda: {})


@dataclass
class FileResponse:
    file: Optional[BytesIO]
    name: str
    mimetype: str


def build_json_response(status: int, body: Dict[str, Any]):
    return Response(status, json.dumps(body), headers={"content-type": "application/json"})


def build_client_error_response(reasons: List[str]) -> Response:
    return build_json_response(
        400,
        {"errors": reasons},
    )


def build_success_response(body: Dict[str, Any]):
    return build_json_response(200, body)


def build_unknown_response():
    return Response(404)


def build_no_content_response():
    return Response(204)


def build_method_not_allowed_response():
    return Response(405)



import sys

if sys.version_info >= (3, 9):
    import zoneinfo
else:
    from backports import zoneinfo

from datetime import datetime, timedelta
from typing import Any, Dict, Tuple

from jose import jwt


def build_jwt(body: Dict[str, Any], secret: str, expiration: int = 1) -> Tuple[str, Dict[str, Any]]:
    body["exp"] = datetime.timestamp(
        datetime.utcnow().replace(tzinfo=zoneinfo.ZoneInfo("UTC")) + timedelta(hours=expiration)  # type: ignore
    )
    return jwt.encode(body, secret, algorithm="HS256"), body  # type: ignore



from typing import Any, Dict, List, cast

from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerActionField
from forestadmin.agent_toolkit.utils.id import unpack_id
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.exceptions import DatasourceToolkitException
from forestadmin.datasource_toolkit.interfaces.actions import ActionField, ActionFieldType
from forestadmin.datasource_toolkit.interfaces.records import CompositeIdAlias


class ForestValueConverterException(DatasourceToolkitException):
    pass


class ForestValueConverter:
    @staticmethod
    def make_form_data_from_fields(
        datasource: Datasource[Collection], fields: List[ForestServerActionField]
    ) -> Dict[str, Any]:

        form_data: Dict[str, Any] = {}
        for field in fields:
            if field["reference"] and field["value"]:
                collection_name = field["reference"].split(":")[0]
                collection = datasource.get_collection(collection_name)
                form_data[field["field"]] = unpack_id(collection.schema, field["value"])
            elif field["type"] == ActionFieldType.FILE:
                pass
            else:
                form_data[field["field"]] = field["value"]
        return form_data

    @staticmethod
    def make_form_unsafe_data(raw_data: Dict[str, Any]) -> Dict[str, Any]:
        return raw_data

    @staticmethod
    def make_form_data(
        datasource: Datasource[Collection], raw_data: Dict[str, Any], fields: List[ActionField]
    ) -> Dict[str, Any]:
        data: Dict[str, Any] = {}
        for key, value in raw_data.items():
            field: ActionField = list(filter(lambda f: f["label"] == key, fields))[0]  # type: ignore
            if field["type"] == ActionFieldType.COLLECTION and value:
                collection = datasource.get_collection(cast(str, field["collection_name"]))
                data[key] = unpack_id(collection.schema, value)
            elif field["type"] == ActionFieldType.FILE:
                data[key] = ""
            else:
                data[key] = value
        return data

    @staticmethod
    def value_to_forest(field: ActionField, value: Any) -> Any:
        if field["type"] == ActionFieldType.ENUM:
            if not field["enum_values"] or ((value is not None) and value not in field["enum_values"]):
                raise ForestValueConverterException(f"{value} is not in {field['enum_values']}")
            return value

        elif field["type"] == ActionFieldType.ENUM_LIST:
            for v in cast(List[str], value):
                if not field["enum_values"] or v not in field["enum_values"]:
                    raise ForestValueConverterException(f"{v} is not in {field['enum_values']}")
            return value
        elif field["type"] == ActionFieldType.COLLECTION:
            value = cast(CompositeIdAlias, value)
            return "|".join(value)

        return value



import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import Any, Dict, Optional

import aiohttp
from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.options import Options


class ForestHttpApiException(AgentToolkitException):
    pass


class HttpOptions(TypedDict):
    env_secret: str
    forest_server_url: str
    is_production: bool


class ForestHttpApi:
    @staticmethod
    def build_enpoint(forest_server_url: str, url: str):
        return f"{forest_server_url}{url}"

    @classmethod
    async def get_open_id_issuer_metadata(cls, option: Options) -> Dict[str, Any]:
        endpoint = cls.build_enpoint(option["forest_server_url"], "/oidc/.well-known/openid-configuration")
        response = await cls.get(endpoint, {"forest-secret-key": option["env_secret"]})
        return response

    @classmethod
    async def get_rendering_authorization(cls, rendering_id: int, access_token: str, option: Options):
        endpoint = cls.build_enpoint(option["forest_server_url"], f"/liana/v2/renderings/{rendering_id}/authorization")
        response = await cls.get(
            endpoint,
            {
                "forest-token": access_token,
                "forest-secret-key": option["env_secret"],
            },
        )
        return response

    @classmethod
    async def get_permissions(cls, option: HttpOptions, rendering_id: int) -> Dict[str, Any]:  # type: ignore
        endpoint = cls.build_enpoint(option["forest_server_url"], f"/liana/v3/permissions?renderingId={rendering_id}")
        headers = {"forest-secret-key": option["env_secret"]}
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(endpoint, headers=headers) as response:
                    if response.status == 200:
                        return await response.json()
            except aiohttp.ClientError:
                raise ForestHttpApiException(f"Failed to fetch {endpoint}")

    @staticmethod
    async def get(endpoint: str, headers: Dict[str, str]) -> Dict[str, Any]:
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(endpoint, headers=headers) as response:
                    if response.status == 200:
                        return await response.json()
                    raise ForestHttpApiException(
                        f"Failed to fetch {endpoint} ({response.status}, {headers} {await response.text()})"
                    )
            except aiohttp.ClientError:
                raise ForestHttpApiException(f"Failed to fetch {endpoint}")

    @staticmethod
    async def post(endpoint: str, body: Dict[str, Any], headers: Dict[str, str]) -> Optional[Dict[str, Any]]:
        async with aiohttp.ClientSession() as session:
            try:
                async with session.post(endpoint, json=body, headers=headers) as response:
                    print("api map upload status", response.status)
                    if response.status == 200:
                        return await response.json()
                    return None
            except aiohttp.ClientError:
                raise ForestHttpApiException(f"Failed to fetch {endpoint}")



from typing import Any, Dict, Optional, Union

from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.resources.collections import BoundCollection
from forestadmin.agent_toolkit.resources.collections.requests import RequestArgs as CollectionRequestArgs
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection
from forestadmin.agent_toolkit.utils.context import Request, RequestMethod, User
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.decorators.datasource import CustomizedDatasource
from typing_extensions import Self


class RequestActionException(AgentToolkitException):
    pass


class RequestArgs(CollectionRequestArgs):
    action_name: str


class ActionRequest(RequestCollection):
    def __init__(
        self,
        method: RequestMethod,
        action_name: str,
        collection: Union[Collection, CustomizedCollection],
        body: Optional[Dict[str, Any]] = None,
        query: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        user: Optional[User] = None,
    ):
        super(ActionRequest, self).__init__(method, collection, body, query, headers, user)
        self.action_name = action_name

    @staticmethod
    def from_request_args(
        request: Request, datasource: Union[Datasource[BoundCollection], CustomizedDatasource]
    ) -> RequestArgs:
        request_args = RequestCollection.from_request_args(request, datasource)
        try:
            action_name: int = request.query["action_name"]  # type: ignore
        except KeyError:
            raise RequestActionException("'action_name' is missing in the request")
        name = list(request_args["collection"].schema["actions"].keys())[action_name]
        return {**request_args, "action_name": name}

    @classmethod
    def from_request(
        cls, request: Request, datasource: Union[Datasource[BoundCollection], CustomizedDatasource]
    ) -> Self:
        return cls(**cls.from_request_args(request, datasource))



from typing import List, Optional, Set

from forestadmin.datasource_toolkit.interfaces.fields import ColumnAlias, Operator, PrimitiveType


class FrontendFilterableUtils:
    BASE_OPERATORS = [Operator.EQUAL, Operator.NOT_EQUAL, Operator.PRESENT, Operator.BLANK]

    DATE_OPERATORS = [
        Operator.LESS_THAN,
        Operator.GREATER_THAN,
        Operator.TODAY,
        Operator.YESTERDAY,
        Operator.PREVIOUS_X_DAYS,
        Operator.PREVIOUS_WEEK,
        Operator.PREVIOUS_QUARTER,
        Operator.PREVIOUS_MONTH,
        Operator.PREVIOUS_YEAR,
        Operator.PREVIOUS_X_DAYS_TO_DATE,
        Operator.PREVIOUS_WEEK_TO_DATE,
        Operator.PREVIOUS_QUARTER_TO_DATE,
        Operator.PREVIOUS_MONTH_TO_DATE,
        Operator.PREVIOUS_YEAR_TO_DATE,
        Operator.PAST,
        Operator.FUTURE,
        Operator.BEFORE_X_HOURS_AGO,
        Operator.AFTER_X_HOURS_AGO,
    ]

    OPERATOR_BY_TYPES = {
        PrimitiveType.BOOLEAN: BASE_OPERATORS,
        PrimitiveType.ENUM: [*BASE_OPERATORS, Operator.IN],
        PrimitiveType.NUMBER: [*BASE_OPERATORS, Operator.IN, Operator.GREATER_THAN, Operator.LESS_THAN],
        PrimitiveType.STRING: [
            *BASE_OPERATORS,
            Operator.IN,
            Operator.STARTS_WITH,
            Operator.ENDS_WITH,
            Operator.CONTAINS,
            Operator.NOT_CONTAINS,
        ],
        PrimitiveType.UUID: BASE_OPERATORS,
        PrimitiveType.DATE: [*BASE_OPERATORS, *DATE_OPERATORS],
        PrimitiveType.DATE_ONLY: [*BASE_OPERATORS, *DATE_OPERATORS],
        PrimitiveType.TIME_ONLY: [*BASE_OPERATORS, *DATE_OPERATORS],
    }

    @classmethod
    def is_filterable(cls, column_type: ColumnAlias, operators: Optional[Set[Operator]]) -> bool:
        required_operators = cls.get_required_operators(column_type) or []
        return bool(operators and all([op in operators for op in required_operators]))

    @classmethod
    def get_required_operators(cls, column_type: ColumnAlias) -> Optional[List[Operator]]:
        res = None
        if isinstance(column_type, PrimitiveType):
            res = cls.OPERATOR_BY_TYPES.get(column_type)
        elif isinstance(column_type, list):
            res = [Operator.INCLUDES_ALL]
        return res



from typing import List, Union

from forestadmin.agent_toolkit.utils.forest_schema.generator_action import SchemaActionGenerator
from forestadmin.agent_toolkit.utils.forest_schema.generator_field import SchemaFieldGenerator
from forestadmin.agent_toolkit.utils.forest_schema.generator_segment import SchemaSegmentGenerator
from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerCollection, ForestServerField
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.fields import FieldType
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils


class SchemaCollectionGenerator:
    @staticmethod
    async def build(prefix: str, collection: Union[Collection, CustomizedCollection]) -> ForestServerCollection:
        fields: List[ForestServerField] = []
        for field_name in collection.schema["fields"].keys():
            if not SchemaUtils.is_foreign_key(collection.schema, field_name):
                fields.append(SchemaFieldGenerator.build(collection, field_name))

        return {
            "actions": [
                await SchemaActionGenerator.build(prefix, collection, name)
                for name in collection.schema["actions"].keys()
            ],
            "fields": fields,
            "icon": None,
            "integration": None,
            "isReadOnly": all(
                [f["type"] == FieldType.COLUMN and f["is_read_only"] for f in collection.schema["fields"].values()]
            ),
            "isSearchable": collection.schema["searchable"],
            "isVirtual": False,
            "name": collection.name,
            "onlyForRelationships": False,
            "paginationType": "page",
            "segments": [
                await SchemaSegmentGenerator.build(collection, name) for name in collection.schema["segments"]
            ],
        }



import sys

from forestadmin.agent_toolkit.resources.collections.stats import StatsResource

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.options import AgentMeta, Options
from forestadmin.agent_toolkit.resources.actions.resources import ActionResource
from forestadmin.agent_toolkit.resources.collections import BoundCollection
from forestadmin.agent_toolkit.resources.collections.crud import CrudResource
from forestadmin.agent_toolkit.resources.collections.crud_related import CrudRelatedResource
from forestadmin.agent_toolkit.resources.security.resources import Authentication
from forestadmin.agent_toolkit.services.permissions import PermissionService
from forestadmin.agent_toolkit.services.serializers.json_api import create_json_api_schema
from forestadmin.agent_toolkit.utils.forest_schema.emitter import SchemaEmitter
from forestadmin.agent_toolkit.utils.http import ForestHttpApi
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.decorators.datasource import CustomizedDatasource


class Resources(TypedDict):
    authentication: Authentication
    crud: CrudResource
    crud_related: CrudRelatedResource
    stats: StatsResource
    actions: ActionResource


class Agent:
    def __init__(self, options: Options):
        self.options = options
        self.composite_datasource: CustomizedDatasource[CustomizedCollection] = Datasource()  # type: ignore

        self.permission_service = PermissionService(
            {
                "env_secret": options["env_secret"],
                "forest_server_url": options["forest_server_url"],
                "is_production": options["is_production"],
                "permission_cache_duration": 60,
            }
        )

    @property
    def resources(self) -> Resources:
        return {
            "authentication": Authentication(self.options),
            "crud": CrudResource(self.composite_datasource, self.permission_service, self.options),
            "crud_related": CrudRelatedResource(self.composite_datasource, self.permission_service, self.options),
            "stats": StatsResource(self.composite_datasource, self.permission_service, self.options),
            "actions": ActionResource(self.composite_datasource, self.permission_service, self.options),
        }

    def add_datasource(self, datasource: Datasource[BoundCollection]):
        customized_datasource = CustomizedDatasource(datasource)
        for collection in datasource.collections:
            collection = CustomizedCollection(collection, customized_datasource)
            self.composite_datasource.add_collection(collection)
            customized_datasource.add_collection(collection)

        for collection in customized_datasource.collections:
            # second loop is mandatory to have all collection set
            create_json_api_schema(collection)

    @property
    def meta(self) -> AgentMeta:
        try:
            return getattr(self, "META")
        except AttributeError:
            raise AgentToolkitException("The agent subclass should set the META attribute")

    async def start(self):
        collections_schema = await SchemaEmitter.get_serialized_schema(
            self.options, self.composite_datasource, self.meta
        )
        api_map = SchemaEmitter.serialize(collections_schema, self.meta)
        import json

        print(json.dumps(api_map))
        await ForestHttpApi.post(
            ForestHttpApi.build_enpoint(self.options["forest_server_url"], "/forest/apimaps"),
            api_map,
            {"forest-secret-key": self.options["env_secret"], "content-type": "application/json"},
        )



import sys

if sys.version_info >= (3, 9):
    import zoneinfo
else:
    from backports import zoneinfo

import json
from typing import Any, Dict, List, Optional, Tuple, Union

from forestadmin.agent_toolkit.exceptions import AgentToolkitException
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection, RequestRelationCollection
from forestadmin.agent_toolkit.utils.context import Request
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.fields import is_column, is_many_to_one, is_one_to_one
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.factory import (
    ConditionTreeFactory,
    ConditionTreeFactoryException,
)
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter, FilterComponent
from forestadmin.datasource_toolkit.interfaces.query.page import Page
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.query.projections.factory import ProjectionFactory
from forestadmin.datasource_toolkit.interfaces.query.sort import Sort
from forestadmin.datasource_toolkit.interfaces.query.sort.factory import SortFactory
from forestadmin.datasource_toolkit.interfaces.records import CompositeIdAlias
from forestadmin.datasource_toolkit.validations.condition_tree import (
    ConditionTreeValidator,
    ConditionTreeValidatorException,
)
from forestadmin.datasource_toolkit.validations.projection import ProjectionValidator

DEFAULT_ITEMS_PER_PAGE = 15
DEFAULT_PAGE_TO_SKIP = 1


class FilterException(AgentToolkitException):
    pass


def _all_records_subset_query(request: Request) -> Dict[str, Any]:
    try:
        return request.body.get("data", {}).get("attributes", {}).get("all_records_subset_query", {})  # type: ignore
    except AttributeError:
        # data may be a list
        return {}


def _subset_or_query(request: Request, key: str) -> Optional[str]:
    res = None
    if request.body:
        sub_res = _all_records_subset_query(request).get(key)
        if sub_res:
            res = str(sub_res)
    if not res and request.query:
        res = request.query.get(key)
        if res:
            res = str(res)
    return res


def parse_selection_ids(request: Request) -> Tuple[List[CompositeIdAlias], bool]:
    if request.body:
        try:
            attributes: Dict[str, Any] = request.body.get("data", {}).get("attributes", {})  # type: ignore
        except AttributeError:
            attributes = {}
        exclude_ids = bool(attributes.get("all_records", False))  # type: ignore
        if "ids" in attributes:
            ids = [[id] for id in attributes["ids"]]
        elif isinstance(request.body.get("data"), list):
            ids = [[r["id"]] for r in request.body["data"]]
        else:
            ids = []
        return ids, exclude_ids

    raise Exception()


def _get_collection(
    request: Union[RequestCollection, RequestRelationCollection]
) -> Union[CustomizedCollection, Collection]:
    collection = request.collection
    if isinstance(request, RequestRelationCollection):
        collection = request.foreign_collection
    return collection


def parse_sort(request: Union[RequestCollection, RequestRelationCollection]):
    sort_string: Optional[str] = _subset_or_query(request, "sort")
    if not sort_string:
        return SortFactory.by_primary_keys(_get_collection(request))

    sort_field = sort_string
    is_descending = sort_string[0] == "-"
    if is_descending:
        sort_field = sort_string[1:]
    return Sort([{"field": sort_field, "ascending": not is_descending}])


def parse_page(request: Request) -> Page:
    items_per_page: Optional[int] = None
    page_to_skip: Optional[int] = None
    if request.body:
        subset_query = _all_records_subset_query(request)
        if "page[size]" in subset_query:
            items_per_page = subset_query["page[size]"]
        if "page[number]" in subset_query:
            page_to_skip = subset_query["page[number]"]

    if not items_per_page and not page_to_skip and request.query:
        if "page[size]" in request.query:
            items_per_page = int(request.query["page[size]"])
        if "page[number]" in request.query:
            page_to_skip = int(request.query["page[number]"])

    if not items_per_page:
        items_per_page = DEFAULT_ITEMS_PER_PAGE
    if not page_to_skip:
        page_to_skip = DEFAULT_PAGE_TO_SKIP

    if (
        not isinstance(items_per_page, int)  # type: ignore
        or not isinstance(page_to_skip, int)  # type: ignore
        or page_to_skip <= 0  # type: ignore
        or items_per_page <= 0
    ):
        raise FilterException(f"Invalid pagination [limit: {items_per_page}, skip: {page_to_skip}]")

    return Page((page_to_skip - 1) * items_per_page, items_per_page)


def parse_search(request: Union[RequestCollection, RequestRelationCollection]):
    search: Optional[str] = _subset_or_query(request, "search")
    if search and not _get_collection(request).schema["searchable"]:
        raise FilterException("Collection is not searchable")

    return search


def parse_search_extended(request: Request) -> bool:
    extended: Optional[str] = _subset_or_query(request, "searchExtended")
    return extended is not None and extended not in ["0", "false"]


def parse_segment(request: Union[RequestCollection, RequestRelationCollection]) -> Optional[str]:
    segment: Optional[str] = _subset_or_query(request, "segment")
    if segment and segment not in _get_collection(request).schema["segments"]:
        raise FilterException(f"Invalid segment {segment}")
    return segment


def parse_timezone(request: Request) -> zoneinfo.ZoneInfo:
    if not request.query or "timezone" not in request.query:
        raise FilterException("Missing timezone")

    tz = request.query["timezone"]
    try:
        return zoneinfo.ZoneInfo(tz)
    except zoneinfo.ZoneInfoNotFoundError:
        raise FilterException(f"Invalid timezone {tz}")


def parse_condition_tree(request: Union[RequestCollection, RequestRelationCollection]) -> Optional[ConditionTree]:
    filters: Optional[str] = _subset_or_query(request, "filters")
    if not filters and request.body and "filters" in request.body:
        filters = request.body["filters"]
    if not filters:
        return None

    jsoned_filters = json.loads(filters)
    try:
        condition_tree = ConditionTreeFactory.from_plain_object(jsoned_filters)
    except ConditionTreeFactoryException as e:
        raise FilterException(str(e))

    try:
        ConditionTreeValidator.validate(condition_tree, _get_collection(request))
    except ConditionTreeValidatorException as e:
        raise FilterException(str(e))

    return condition_tree


def _parse_projection_fields(
    query: Dict[str, Any],
    collection: Union[CustomizedCollection, Collection],
    front_collection_name: str,
    is_related: bool = False,
) -> List[str]:
    projection_fields: List[str] = []
    try:
        fields: str = query[f"fields[{front_collection_name}]"]
    except KeyError:
        return ProjectionFactory.all(collection)

    if fields == "":
        return ProjectionFactory.all(collection)
    for field_name in fields.split(","):
        field_schema = collection.get_field(field_name)
        if is_column(field_schema):
            if is_related:
                projection_fields.append(f"{front_collection_name}:{field_name}")
            else:
                projection_fields.append(field_name)
        elif is_one_to_one(field_schema) or is_many_to_one(field_schema):
            fk_collection = collection.datasource.get_collection(field_schema["foreign_collection"])
            projection_fields.extend(_parse_projection_fields(query, fk_collection, field_name, True))
    return projection_fields


def parse_projection(request: Union[RequestCollection, RequestRelationCollection]):
    collection = _get_collection(request)
    if not request.query:
        return ProjectionFactory.all(collection)

    projection_fields = _parse_projection_fields(request.query, collection, collection.name)
    ProjectionValidator.validate(_get_collection(request), projection_fields)
    return Projection(*projection_fields)


def parse_projection_with_pks(request: Union[RequestCollection, RequestRelationCollection]):
    projection = parse_projection(request)
    return projection.with_pks(_get_collection(request))


def build_paginated_filter(
    request: Union[RequestCollection, RequestRelationCollection], scope: Optional[ConditionTree]
) -> PaginatedFilter:
    _filter = build_filter(request, scope)
    res = PaginatedFilter(
        {"sort": parse_sort(request), "page": parse_page(request), **_filter.to_filter_component()}  # type: ignore
    )
    return res


def build_filter(request: Union[RequestCollection, RequestRelationCollection], scope: Optional[ConditionTree]):
    filter_component: FilterComponent = {
        "search": parse_search(request),
        "search_extended": parse_search_extended(request),
        "timezone": parse_timezone(request),
    }
    segment = parse_segment(request)
    if segment:
        filter_component["segment"] = segment

    trees: List[ConditionTree] = []
    if scope:
        trees.append(scope)
    condition_tree = parse_condition_tree(request)
    if condition_tree:
        trees.append(condition_tree)

    if trees:
        filter_component["condition_tree"] = ConditionTreeFactory.intersect(trees)

    return Filter(filter_component)



from typing import Dict, List, Union, cast

from forestadmin.agent_toolkit.utils.forest_schema.filterable import FrontendFilterableUtils
from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerField, RelationServer
from forestadmin.agent_toolkit.utils.forest_schema.validation import FrontendValidationUtils
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.fields import (
    Column,
    ColumnAlias,
    FieldType,
    ManyToMany,
    ManyToOne,
    OneToMany,
    OneToOne,
    Operator,
    PrimitiveType,
    RelationAlias,
    is_column,
    is_many_to_many,
    is_many_to_one,
    is_one_to_many,
    is_one_to_one,
)
from forestadmin.datasource_toolkit.utils.collections import CollectionUtils


class SchemaFieldGenerator:
    RELATION_MAPPING: Dict[FieldType, RelationServer] = {
        FieldType.ONE_TO_ONE: "HasOne",
        FieldType.ONE_TO_MANY: "BelongsToMany",
        FieldType.MANY_TO_ONE: "BelongsTo",
        FieldType.MANY_TO_MANY: "BelongsToMany",
    }

    @classmethod
    def build(cls, collection: Union[Collection, CustomizedCollection], field_name: str) -> ForestServerField:
        schema: ForestServerField = {}

        field_schema = collection.get_field(field_name)
        if is_column(field_schema):
            schema = cls.build_column_schema(field_name, field_schema)
        elif (
            is_one_to_one(field_schema)
            or is_one_to_many(field_schema)
            or is_many_to_one(field_schema)
            or is_many_to_many(field_schema)
        ):
            schema = cls.build_relation_schema(collection, field_name, field_schema)
        else:
            raise
        return cast(ForestServerField, dict(sorted(schema.items())))

    @staticmethod
    def build_column_type(_column_type: ColumnAlias) -> ColumnAlias:
        column_type: ColumnAlias
        if isinstance(_column_type, PrimitiveType):
            column_type = _column_type.value
        elif isinstance(_column_type, list):
            _type = cast(List[PrimitiveType], _column_type)
            column_type = [t.value for t in _column_type]
        else:
            _type = cast(Dict[str, PrimitiveType], _column_type)
            column_type = {k: t.value for k, t in _type.items()}
        return column_type

    @classmethod
    def build_column_schema(cls, name: str, column: Column) -> ForestServerField:
        validations = []
        if column["validations"]:
            validations = column["validations"]

        res = {
            "type": cls.build_column_type(column["column_type"]),
            "validations": FrontendValidationUtils.convert_validation_list(column["validations"]),
            "defaultValue": column["default_value"],
            "enums": column["enum_values"],
            "field": name,
            "integration": None,
            "inverseOf": None,
            "isFilterable": FrontendFilterableUtils.is_filterable(column["column_type"], column["filter_operators"]),
            "isPrimaryKey": bool(column["is_primary_key"]),
            "isSortable": bool(column["is_sortable"]),
            "isReadOnly": bool(column["is_read_only"]),
            "isRequired": any([v["operator"] == Operator.PRESENT for v in validations]),
            "isVirtual": False,
            "reference": None,
        }
        return ForestServerField(**res)

    @staticmethod
    def is_foreign_collection_filterable(foreign_collection: Collection) -> bool:
        res = False
        for field in foreign_collection.schema["fields"].values():
            if is_column(field) and FrontendFilterableUtils.is_filterable(
                field["column_type"], field["filter_operators"]
            ):
                res = True
                break
        return res

    @classmethod
    def build_one_to_one_schema(
        cls, relation: OneToOne, foreign_collection: Collection, base_schema: ForestServerField
    ) -> ForestServerField:
        key = relation["origin_key_target"]
        key_schema = cast(Column, foreign_collection.get_field(key))
        return {
            **base_schema,
            "type": cls.build_column_type(key_schema["column_type"]),
            "defaultValue": None,
            "isFilterable": cls.is_foreign_collection_filterable(foreign_collection),
            "isPrimaryKey": False,
            "isRequired": False,
            "isSortable": bool(key_schema["is_sortable"]),
            "validations": [],
            "reference": f"{foreign_collection.name}.{key}",
        }

    @classmethod
    def build_to_many_relation_schema(
        cls, relation: Union[OneToMany, ManyToMany], foreign_collection: Collection, base_schema: ForestServerField
    ) -> ForestServerField:
        if is_one_to_many(relation):
            key = relation["origin_key_target"]
        else:
            relation = cast(ManyToMany, relation)
            key = relation["foreign_key_target"]

        key_schema = cast(Column, foreign_collection.get_field(key))

        return {
            **base_schema,
            "type": [cls.build_column_type(key_schema["column_type"])],
            "defaultValue": None,
            "isFilterable": False,
            "isPrimaryKey": False,
            "isRequired": False,
            "isSortable": False,
            "validations": [],
            "reference": f"{foreign_collection.name}.{key}",
        }

    @classmethod
    def build_many_to_one_schema(
        cls,
        relation: ManyToOne,
        collection: Union[CustomizedCollection, Collection],
        foreign_collection: Collection,
        base_schema: ForestServerField,
    ) -> ForestServerField:
        key = relation["foreign_key"]
        key_schema = cast(Column, collection.get_field(key))
        validations = key_schema["validations"] or []

        return {
            **base_schema,
            "type": cls.build_column_type(key_schema["column_type"]),
            "defaultValue": key_schema["default_value"],
            "isFilterable": cls.is_foreign_collection_filterable(foreign_collection),
            "isPrimaryKey": bool(key_schema["is_primary_key"]),
            "isRequired": any([v["operator"] == Operator.PRESENT for v in validations]),
            "isSortable": bool(key_schema["is_sortable"]),
            "validations": FrontendValidationUtils.convert_validation_list(validations),
            "reference": f"{foreign_collection.name}.{relation['foreign_key_target']}",
        }

    @classmethod
    def build_relation_schema(
        cls, collection: Union[Collection, CustomizedCollection], field_name: str, field_schema: RelationAlias
    ) -> ForestServerField:
        foreign_collection = collection.datasource.get_collection(field_schema["foreign_collection"])
        relation_schema: ForestServerField = {
            "field": field_name,
            "enums": None,
            "isReadOnly": False,
            "isVirtual": False,
            "inverseOf": CollectionUtils.get_inverse_relation(cast(Collection, collection), field_name),
            "relationship": cls.RELATION_MAPPING[field_schema["type"]],
        }
        if is_many_to_many(field_schema) or is_one_to_many(field_schema):
            res = cls.build_to_many_relation_schema(field_schema, foreign_collection, relation_schema)
        elif is_one_to_one(field_schema):
            res = cls.build_one_to_one_schema(field_schema, foreign_collection, relation_schema)
        else:
            res = cls.build_many_to_one_schema(
                cast(ManyToOne, field_schema), collection, foreign_collection, relation_schema
            )
        return res



import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal
if sys.version_info >= (3, 9):
    import zoneinfo
else:
    from backports import zoneinfo

from typing import Any, Dict, List, Union, cast

from forestadmin.agent_toolkit.resources.collections import BaseCollectionResource
from forestadmin.agent_toolkit.resources.collections.decorators import authenticate, authorize, check_method
from forestadmin.agent_toolkit.resources.collections.exceptions import CollectionResourceException
from forestadmin.agent_toolkit.resources.collections.filter import (
    FilterException,
    build_filter,
    build_paginated_filter,
    parse_condition_tree,
    parse_projection_with_pks,
    parse_selection_ids,
    parse_timezone,
)
from forestadmin.agent_toolkit.resources.collections.requests import (
    RequestCollectionException,
    RequestRelationCollection,
)
from forestadmin.agent_toolkit.services.serializers import DumpedResult, add_search_metadata
from forestadmin.agent_toolkit.services.serializers.json_api import JsonApiException, JsonApiSerializer
from forestadmin.agent_toolkit.utils.context import (
    Request,
    RequestMethod,
    Response,
    build_client_error_response,
    build_no_content_response,
    build_success_response,
)
from forestadmin.agent_toolkit.utils.id import unpack_id
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import DatasourceException
from forestadmin.datasource_toolkit.interfaces.fields import (
    Operator,
    is_many_to_many,
    is_many_to_one,
    is_one_to_many,
    is_one_to_one,
)
from forestadmin.datasource_toolkit.interfaces.query.aggregation import Aggregation
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.factory import ConditionTreeFactory
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.filter.factory import FilterFactory
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.records import CompositeIdAlias
from forestadmin.datasource_toolkit.utils.collections import CollectionUtils
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils
from forestadmin.datasource_toolkit.validations.field import FieldValidatorException

LiteralMethod = Literal["list", "add", "delete", "update"]


class CrudRelatedResource(BaseCollectionResource):
    async def dispatch(self, request: Request, method_name: LiteralMethod) -> Response:
        method = getattr(self, method_name)
        try:
            request_collection = RequestRelationCollection.from_request(request, self.datasource)
        except RequestCollectionException as e:
            return build_client_error_response([str(e)])
        return await method(request_collection)

    @authenticate
    @authorize("browse")
    @check_method(RequestMethod.GET)
    async def list(self, request: RequestRelationCollection) -> Response:
        if not (is_one_to_many(request.relation) or is_many_to_many(request.relation)):
            return build_client_error_response(["Unhandled relation type"])
        try:
            ids = unpack_id(request.collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])
        scope_tree = await self.permission.get_scope(request, request.foreign_collection)
        paginated_filter = build_paginated_filter(request, scope_tree)
        projection = parse_projection_with_pks(request)
        records = await CollectionUtils.list_relation(
            cast(Collection, request.collection),
            ids,
            cast(Collection, request.foreign_collection),
            request.relation,
            paginated_filter,
            projection,
        )
        schema = JsonApiSerializer.get(request.foreign_collection)
        try:
            dumped: DumpedResult = schema(projections=projection).dump(records, many=True)  # type: ignore
        except JsonApiException as e:
            return build_client_error_response([str(e)])

        if paginated_filter.search:
            dumped = add_search_metadata(dumped, paginated_filter.search)

        return build_success_response(cast(Dict[str, Any], dumped))

    @authenticate
    @authorize("edit")
    @check_method(RequestMethod.POST)
    async def add(self, request: RequestRelationCollection) -> Response:
        try:
            parent_ids = unpack_id(request.collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        if (
            not request.body
            or "data" not in request.body
            or len(request.body["data"]) == 0
            or not request.body["data"][0].get("id")
        ):
            return build_client_error_response(["missing target's id "])
        try:
            targeted_relation_ids = unpack_id(request.foreign_collection.schema, request.body["data"][0]["id"])
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        if not (is_one_to_many(request.relation) or is_many_to_many(request.relation)):
            return build_client_error_response(["Unhandled relation type"])

        pks = SchemaUtils.get_primary_keys(request.foreign_collection.schema)[0]
        value = await CollectionUtils.get_value(
            cast(Collection, request.foreign_collection), targeted_relation_ids, pks
        )
        if is_one_to_many(request.relation):
            return await self._associate_one_to_many(request, parent_ids, value)
        else:
            return await self._associate_many_to_many(request, parent_ids, value)

    @authenticate
    @authorize("edit")
    @check_method(RequestMethod.PUT)
    async def update_list(self, request: RequestRelationCollection) -> Response:
        try:
            timezone = parse_timezone(request)
        except FilterException as e:
            return build_client_error_response([str(e)])
        try:
            parent_id = unpack_id(request.collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])
        if not request.body or not request.body.get("data") or "id" not in request.body["data"]:
            return build_client_error_response(["Relation id is missing"])
        try:
            linked_id = unpack_id(request.foreign_collection.schema, request.body["data"]["id"])
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        if is_many_to_one(request.relation) or is_one_to_one(request.relation):

            if is_many_to_one(request.relation):
                meth = self._update_many_to_one
            else:
                meth = self._update_one_to_one

            try:
                await meth(request, parent_id, linked_id, timezone)
            except (CollectionResourceException, DatasourceException) as e:
                return build_client_error_response([str(e)])
            return build_no_content_response()
        return build_client_error_response(["Unhandled relation type"])

    @authenticate
    @authorize("browse")
    @check_method(RequestMethod.GET)
    async def count(self, request: RequestRelationCollection) -> Response:
        try:
            parent_id = unpack_id(request.collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])
        scope_tree = await self.permission.get_scope(request, request.foreign_collection)
        filter = build_filter(request, scope_tree)
        aggregation = Aggregation({"operation": "Count"})

        result = await CollectionUtils.aggregate_relation(
            cast(Collection, request.collection), parent_id, request.relation_name, filter, aggregation
        )
        try:
            count = result[0]["value"]
        except IndexError:
            count = 0
        return build_success_response({"count": count})

    @authenticate
    @authorize("delete")
    @check_method(RequestMethod.DELETE)
    async def delete_list(self, request: RequestRelationCollection) -> Response:
        try:
            parent_ids = unpack_id(request.collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        delete_mode = False
        if request.query:
            delete_mode = bool(request.query.get("delete", False))

        filter = await self.get_base_fk_filter(request)

        if is_one_to_many(request.relation) or is_many_to_many(request.relation):
            if is_one_to_many(request.relation):
                meth = self._delete_one_to_many
            else:
                meth = self._delete_many_to_many
            try:
                await meth(request, parent_ids, delete_mode, filter)
            except (DatasourceException, CollectionResourceException) as e:
                return build_client_error_response([str(e)])
            return build_no_content_response()
        return build_client_error_response(["Unhandled relation type"])

    async def _associate_one_to_many(
        self, request: RequestRelationCollection, parent_ids: CompositeIdAlias, id_value: Union[str, int]
    ) -> Response:
        if not is_one_to_many(request.relation):
            raise
        scope_tree = await self.permission.get_scope(request, request.foreign_collection)
        filter = build_filter(request, scope_tree)
        trees: List[ConditionTree] = [
            ConditionTreeLeaf("id", Operator.EQUAL, id_value),
        ]
        if filter.condition_tree:
            trees.append(filter.condition_tree)
        filter.condition_tree = ConditionTreeFactory.intersect(trees)
        value = await CollectionUtils.get_value(
            cast(Collection, request.collection), parent_ids, request.relation["origin_key_target"]
        )
        try:
            await request.foreign_collection.update(filter, {f"{request.relation['origin_key']}": value})
        except DatasourceException as e:
            return build_client_error_response([str(e)])
        else:
            return build_no_content_response()

    async def _associate_many_to_many(
        self, request: RequestRelationCollection, parent_ids: CompositeIdAlias, foreign_id_value: Union[str, int]
    ):
        if not is_many_to_many(request.relation):
            raise
        id = SchemaUtils.get_primary_keys(request.collection.schema)
        origin_id_value = await CollectionUtils.get_value(cast(Collection, request.collection), parent_ids, id[0])

        record = {
            f"{request.relation['origin_key']}": origin_id_value,
            f"{request.relation['foreign_key']}": foreign_id_value,
        }
        through_collection = request.collection.datasource.get_collection(request.relation["through_collection"])
        try:
            await through_collection.create([record])
        except DatasourceException as e:
            return build_client_error_response([str(e)])
        else:
            return build_no_content_response()

    async def get_base_fk_filter(self, request: RequestRelationCollection):
        ids, exclude_ids = parse_selection_ids(request)

        if len(ids) == 0 and not exclude_ids:
            raise

        selected_ids = ConditionTreeFactory.match_ids(request.foreign_collection.schema, ids)
        if exclude_ids:
            selected_ids = selected_ids.inverse()

        scope_tree = await self.permission.get_scope(request, request.foreign_collection)
        filter = build_filter(request, scope_tree)
        trees = [selected_ids]
        if scope_tree:
            trees.append(scope_tree)
        parsed_tree = parse_condition_tree(request)
        if parsed_tree:
            trees.append(parsed_tree)
        filter.condition_tree = ConditionTreeFactory.intersect(trees)
        return filter

    async def _delete_one_to_many(
        self,
        request: RequestRelationCollection,
        parent_id: CompositeIdAlias,
        is_delete: bool,
        base_target_filter: Filter,
    ):
        if not is_one_to_many(request.relation):
            raise CollectionResourceException("Unhandled relation type")
        foreign_paginated_filter = await FilterFactory.make_foreign_filter(
            cast(Collection, request.collection), parent_id, request.relation, base_target_filter
        )
        foreign_filter = foreign_paginated_filter.to_base_filter()
        if is_delete:
            await request.foreign_collection.delete(foreign_filter)
        else:
            await request.foreign_collection.update(foreign_filter, {f"{request.relation['origin_key']}": None})

    async def _delete_many_to_many(
        self,
        request: RequestRelationCollection,
        parent_id: CompositeIdAlias,
        is_delete: bool,
        base_target_filter: Filter,
    ):
        if not is_many_to_many(request.relation):
            raise CollectionResourceException("Unhandled relation type")
        through_filter = await FilterFactory.make_through_filter(
            cast(Collection, request.collection),
            parent_id,
            request.relation,
            base_target_filter,
        )
        through_collection = request.collection.datasource.get_collection(request.relation["through_collection"])
        if is_delete:
            foreign_filter = await FilterFactory.make_foreign_filter(
                cast(Collection, request.collection), parent_id, request.relation, base_target_filter
            )
            await through_collection.delete(through_filter.to_base_filter())
            try:
                await request.foreign_collection.delete(foreign_filter.to_base_filter())
            except DatasourceException:
                # Let the datasource crash when:
                # - the records in the foreignCollection are linked to other records in the origin collection
                # - the underlying database/api is not cascading deletes
                pass
        else:
            await through_collection.delete(through_filter.to_base_filter())

    async def _update_one_to_one(
        self,
        request: RequestRelationCollection,
        parent_id: CompositeIdAlias,
        linked_id: CompositeIdAlias,
        timezone: zoneinfo.ZoneInfo,
    ):
        if not is_one_to_one(request.relation):
            raise CollectionResourceException("Unhandled relation type")

        scope = await self.permission.get_scope(request)
        origin_value = await CollectionUtils.get_value(
            cast(Collection, request.collection), parent_id, request.relation["origin_key_target"]
        )

        trees: List[ConditionTree] = [ConditionTreeLeaf(request.relation["origin_key"], Operator.EQUAL, origin_value)]
        if scope:
            trees.append(scope)

        await request.foreign_collection.update(
            Filter({"condition_tree": ConditionTreeFactory.intersect(trees), "timezone": timezone}),
            {f"{request.relation['origin_key']}": origin_value},
        )

        trees = [ConditionTreeFactory.match_ids(request.foreign_collection.schema, [linked_id])]
        if scope:
            trees.append(scope)
        await request.foreign_collection.update(
            Filter(
                {
                    "condition_tree": ConditionTreeFactory.intersect(trees),
                    "timezone": timezone,
                }
            ),
            {f"{request.relation['origin_key']}": origin_value},
        )

    async def _update_many_to_one(
        self,
        request: RequestRelationCollection,
        parent_id: CompositeIdAlias,
        linked_id: CompositeIdAlias,
        timezone: zoneinfo.ZoneInfo,
    ):
        if not is_many_to_one(request.relation):
            raise CollectionResourceException("Unhandled relation type")

        scope = await self.permission.get_scope(request)
        foreign_value = await CollectionUtils.get_value(
            cast(Collection, request.collection), linked_id, request.relation["foreign_key_target"]
        )

        trees = [ConditionTreeFactory.match_ids(request.collection.schema, [parent_id])]
        if scope:
            trees.append(scope)

        await request.collection.update(
            Filter({"condition_tree": ConditionTreeFactory.intersect(trees), "timezone": timezone}),
            {f"{request.relation['foreign_key']}": foreign_value},
        )



import itertools
import json
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

from cachetools import TTLCache
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection, RequestRelationCollection
from forestadmin.agent_toolkit.services.permissions.options import RoleOptions
from forestadmin.agent_toolkit.utils.context import Response, User
from forestadmin.agent_toolkit.utils.http import ForestHttpApi
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.factory import ConditionTreeFactory
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf


class PermissionBody(TypedDict):
    actions: Set[str]
    actions_by_user: Dict[str, Set[int]]
    scopes: Dict[str, Any]


def decode_permission_body(rendering_id: int, body: Dict[str, Any]) -> PermissionBody:
    if not body["meta"].get("rolesACLActivated"):
        raise PermissionServiceException("Roles V2 are unsupported")

    collections: Dict[str, Any] = {}
    stats: Dict[str, Any] = body.get("stats", {})
    renderings: Dict[str, Any] = {}

    if "data" in body:
        if body["data"].get("collections"):
            collections = body["data"]["collections"]

        renderings = body["data"].get("renderings", {})
        if str(rendering_id) in renderings:
            renderings = renderings[str(rendering_id)]
    actions, actions_by_user = _decode_action_permissions(collections)
    actions.update(_decode_chart_permissions(stats))

    scopes = _decode_scopes(renderings)

    return {"actions": actions, "actions_by_user": actions_by_user, "scopes": scopes}


class Scope(TypedDict):
    condition_tree: ConditionTreeBranch
    dynamic_scope_values: Optional[Dict[Any, Any]]


def _decode_scopes(rendering: Dict[str, Any]):
    scopes: Dict[str, Scope] = {}
    for name, v in rendering.items():
        if v.get("scope"):
            scopes[name] = {
                **v["scope"],
                "condition_tree": cast(
                    ConditionTreeBranch, ConditionTreeFactory.from_plain_object(v["scope"]["filter"])
                ),
                "dynamic_scope_values": v["scope"]["dynamicScopesValues"].get("users", {}),
            }
    return scopes


def _decode_action_permissions(collections: Dict[str, Any]) -> Tuple[Set[str], Dict[str, Set[int]]]:
    actions: Set[str] = set()
    actions_by_user: Dict[str, Set[int]] = {}

    for name, settings in collections.items():
        for action_name, user_ids in settings.get("collection", {}).items():
            # Remove 'Enabled' from the name
            short_name = action_name[:-7]
            key = f"{short_name}:{name}"
            if isinstance(user_ids, bool):
                actions.add(key)
            else:
                actions_by_user[key] = set(user_ids)

        for action_name, perms in settings.get("actions", {}).items():
            user_ids = perms["triggerEnabled"]
            key = f"custom:{action_name}:{name}"
            if isinstance(user_ids, bool):
                actions.add(key)
            else:
                actions_by_user[key] = set(user_ids)

    return actions, actions_by_user


def _decode_chart_permissions(stats: Dict[str, Any]):
    server_charts = list(itertools.chain(*stats.values()))
    hashes: List[str] = []
    for chart in server_charts:
        if isinstance(chart, str):  # Queries
            hashes.append(f"chart:{hash(chart)}")
        else:
            frontend_chart = {
                "type": chart.get("type"),
                "filters": chart.get("filter"),
                "aggregate": chart.get("aggregator"),
                "aggregate_field": chart.get("aggregateFieldName"),
                "collection": chart.get("sourceCollectionId"),
                "time_range": chart.get("timeRange"),
                "group_by_date_field": (chart.get("type") == "Line" and chart.get("groupByFieldName")) or None,
                "group_by_field": (chart.get("type") != "Line" and chart.get("groupByFieldName")) or None,
                "limit": chart.get("limit"),
                "label_field": chart.get("labelFieldName"),
                "relationship_field": chart.get("relationshipFieldName"),
            }
            h = hash(json.dumps(dict((k, v) for k, v in frontend_chart.items() if v is not None)))
            hashes.append(f"chart:{h}")
    return hashes


class PermissionServiceException(BaseException):
    def __init__(self, message: str, status: int = 403):
        self.STATUS = status
        self.message = message
        super(PermissionServiceException, self).__init__(message)


class PermissionService:
    def __init__(self, options: RoleOptions):
        self.options = options
        self.cache: TTLCache[int, Any] = TTLCache(maxsize=256, ttl=options["permission_cache_duration"])

    async def _get_rendering_permissions(self, rendering_id: int) -> PermissionBody:
        if rendering_id not in self.cache:
            permission_body = await ForestHttpApi.get_permissions(
                {
                    "env_secret": self.options["env_secret"],
                    "forest_server_url": self.options["forest_server_url"],
                    "is_production": self.options["is_production"],
                },
                rendering_id,
            )
            self.cache[rendering_id] = decode_permission_body(rendering_id, permission_body)
        return self.cache[rendering_id]

    def _invalidate_cache(self, rendering_id: int):
        if rendering_id in self.cache:
            del self.cache[rendering_id]

    async def _can(self, action: str, user: User, allowed_refetch: bool = True) -> Optional[Response]:
        perms = await self._get_rendering_permissions(user.rendering_id)
        is_allowed = action in perms["actions"] or user.user_id in perms["actions_by_user"].get(action, set())
        if not is_allowed and allowed_refetch:
            self._invalidate_cache(user.rendering_id)
            return await self._can(action, user, False)
        elif not is_allowed:
            raise PermissionServiceException("Unauthorized request")

    async def can(self, request: RequestCollection, action: str):
        if not request.user:
            raise PermissionServiceException("Unauthenticated request", 401)
        else:
            await self._can(action, request.user)

    async def get_scope(
        self,
        request: Union[RequestCollection, RequestRelationCollection],
        collection: Union[Collection, CustomizedCollection, None] = None,
    ) -> Optional[ConditionTree]:
        if not request.user:
            raise PermissionServiceException("Unauthenticated request", 401)
        else:
            perms = await self._get_rendering_permissions(request.user.rendering_id)
            name = request.collection.name
            if collection:
                name = collection.name
            try:
                scope: Scope = perms["scopes"][name]
            except KeyError:
                return None

            def build_scope_leaf(tree: ConditionTree) -> ConditionTree:
                if isinstance(tree, ConditionTreeLeaf):
                    dynamic_value = scope["dynamic_scope_values"].get(request.user.user_id)  # type: ignore
                    if isinstance(tree.value, str) and tree.value.startswith("$currentUser"):
                        if dynamic_value:
                            value = dynamic_value[tree.value]
                        elif tree.value.startswith("$currentUser.tags."):
                            value = request.user.tags[tree.value[18:]]  # type: ignore
                        else:
                            value = getattr(request.user, tree.value[13:])

                        return tree.override({"value": value})
                return tree

        return scope["condition_tree"].replace(build_scope_leaf)



import json
from typing import Any, Dict, Optional
from urllib.parse import urljoin

from forestadmin.agent_toolkit.options import Options
from forestadmin.agent_toolkit.utils.http import ForestHttpApi
from oic.oauth2.message import Message
from oic.oic import Client as OicClient
from oic.oic.message import AuthorizationRequest, AuthorizationResponse, ProviderConfigurationResponse


class CustomClientOic(OicClient):
    SCOPE = ["openid", "email", "profile"]

    def register(self, url: str, registration_token: Optional[str] = None, **kwargs):  # type: ignore
        """
        needed to avoid bj64 encoding in authorization header
        """
        req = self.create_registration_request(**kwargs)  # type: ignore
        if self.events:
            self.events.store("Protocol request", req)
        headers = {"content-type": "application/json"}
        if registration_token is not None:
            headers["Authorization"] = "Bearer " + registration_token

        rsp = self.http_request(url, "POST", data=req.to_json(), headers=headers)  # type: ignore

        return self.handle_registration_info(rsp)  # type: ignore

    def get_authorization_url(self, state: str) -> str:
        args: Dict[str, Any] = {
            "client_id": self.client_id,  # type: ignore
            "response_type": "code",
            "scope": self.SCOPE,
            "state": state,
            "redirect_uri": self.registration_response["redirect_uris"][0],
        }
        auth_req: AuthorizationRequest = self.construct_AuthorizationRequest(request_args=args)  # type: ignore
        return auth_req.request(self.authorization_endpoint)  # type: ignore

    def get_parsed_response(self, info: Dict[str, Any]) -> Message:
        return self.parse_response(  # type: ignore
            AuthorizationResponse, info=json.dumps(info), state=info["state"], scope=self.SCOPE
        )

    def get_access_token(self, authorization_response: Message):
        access_token = self.do_access_token_request(  # type: ignore
            state=authorization_response["state"],
            request_args={"code": authorization_response["code"]},
            verify=False,
            skew=5,
            authn_method="",
        )
        return access_token


class ClientFactory:

    oic_client: Optional[CustomClientOic] = None

    @classmethod
    async def build(cls, callback_url: str, options: Options) -> CustomClientOic:
        if cls.oic_client:
            return cls.oic_client

        issuer_metadata: Dict[str, Any] = await ForestHttpApi.get_open_id_issuer_metadata(options)

        client = CustomClientOic()
        client.register(  # type: ignore
            issuer_metadata["registration_endpoint"],
            redirect_uris=[callback_url],
            registration_token=options["env_secret"],
        )
        client.handle_provider_config(
            ProviderConfigurationResponse(  # type: ignore
                jwks_uri=urljoin(issuer_metadata["issuer"], "oidc/jwks"),
                token_endpoint=urljoin(issuer_metadata["issuer"], "oidc/token"),
                authorization_endpoint=urljoin(issuer_metadata["issuer"], "oidc/auth"),
            ),
            issuer_metadata["issuer"],
        )
        cls.oic_client = client
        return cls.oic_client



from abc import abstractmethod
from typing import Any, Union

from forestadmin.agent_toolkit.options import Options
from forestadmin.agent_toolkit.utils.context import FileResponse, Request, Response


class BaseResource:
    def __init__(self, options: Options):
        self.option = options

    @abstractmethod
    async def dispatch(self, request: Request, method_name: Any) -> Union[Response, FileResponse]:
        pass



from typing import Any, List, cast

from forestadmin.datasource_toolkit.collections import CollectionSchema
from forestadmin.datasource_toolkit.interfaces.fields import Column, PrimitiveType
from forestadmin.datasource_toolkit.interfaces.records import CompositeIdAlias, RecordsDataAlias
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils
from forestadmin.datasource_toolkit.validations.field import FieldValidator, FieldValidatorException


class IdException(BaseException):
    pass


def pack_id(schema: CollectionSchema, record: RecordsDataAlias):
    schema_pks = SchemaUtils.get_primary_keys(schema)
    if len(schema_pks) == 0:
        raise IdException("")
    pks = [str(record.get(pk, "")) for pk in schema_pks]
    if not all(pks):
        raise IdException("")

    return "|".join(pks)  # type: ignore


def unpack_id(schema: CollectionSchema, pks: str) -> CompositeIdAlias:

    schema_pks = SchemaUtils.get_primary_keys(schema)
    pk_values = pks.split("|")
    if len(pk_values) != len(schema_pks):
        raise IdException("Unable to unpack the id")

    values: List[Any] = []
    for i, field_name in enumerate(schema_pks):

        schema_field = cast(Column, schema["fields"][field_name])
        value = pk_values[i]

        if schema_field["column_type"] == PrimitiveType.NUMBER:
            try:
                value = int(value)
            except ValueError:
                value = float(value)
        try:
            FieldValidator.validate_value(field_name, schema_field, value)
        except FieldValidatorException:
            raise IdException("Unable to validate the id value")

        values.append(value)

    return values



from datetime import date, datetime
from typing import Any, Dict, List, Literal, Union, cast
from uuid import uuid1

import pandas as pd
from forestadmin.agent_toolkit.resources.collections import BaseCollectionResource
from forestadmin.agent_toolkit.resources.collections.decorators import authenticate, check_method
from forestadmin.agent_toolkit.resources.collections.filter import build_filter
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection, RequestCollectionException
from forestadmin.agent_toolkit.utils.context import (
    FileResponse,
    Request,
    RequestMethod,
    Response,
    build_client_error_response,
    build_success_response,
)
from forestadmin.datasource_toolkit.interfaces.fields import Column, PrimitiveType
from forestadmin.datasource_toolkit.interfaces.query.aggregation import Aggregation
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import Aggregator, ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.transforms.time import Frequency
from forestadmin.datasource_toolkit.interfaces.query.filter.factory import FilterFactory
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils


class StatsResource(BaseCollectionResource):

    FREQUENCIES = {"Day": Frequency.DAY, "Week": Frequency.WEEK, "Month": Frequency.MONTH, "Year": Frequency.YEAR}

    FORMAT = {"Day": "%d/%m/%Y", "Week": "W%W-%Y", "Month": "%m %Y", "Year": "%Y"}

    def stats_method(self, type: str):
        return {
            "Value": self.value,
            "Objective": self.objective,
            "Pie": self.pie,
            "Line": self.line,
            "Leaderboard": self.leader_board,
        }[type]

    async def dispatch(self, request: Request, method_name: Literal["add"]) -> Union[Response, FileResponse]:
        try:
            request_collection = RequestCollection.from_request(request, self.datasource)
        except RequestCollectionException as e:
            return build_client_error_response([str(e)])

        try:
            meth = self.stats_method(request_collection.body["type"])  # type: ignore
        except KeyError:
            return build_client_error_response(
                [f"Unknown stats type {request_collection.body.get('type')}"]  # type: ignore
            )
        except TypeError:
            return build_client_error_response(["Missing stats type in request body"])
        return await meth(request_collection)

    @check_method(RequestMethod.POST)
    @authenticate
    async def value(self, request: RequestCollection) -> Response:
        current_filter = await self._get_gilter(request)
        result = {
            "countCurrent": await self._compute_value(request, current_filter),
        }

        if current_filter.condition_tree:
            current_filter.condition_tree = cast(ConditionTreeBranch, current_filter.condition_tree)
            with_count_previous = self._with_count_previous(current_filter.condition_tree)
            if (
                not hasattr(current_filter.condition_tree, "aggregator")
                or not current_filter.condition_tree.aggregator != Aggregator.AND
            ) and with_count_previous:
                result["countPrevious"] = await self._compute_value(
                    request, FilterFactory.get_previous_period_filter(current_filter)
                )

        return self._build_success_response(result)

    @check_method(RequestMethod.POST)
    @authenticate
    async def objective(self, request: RequestCollection) -> Response:
        current_filter = await self._get_gilter(request)
        result = {"value": await self._compute_value(request, current_filter)}
        return self._build_success_response(result)

    @check_method(RequestMethod.POST)
    @authenticate
    async def pie(self, request: RequestCollection) -> Response:
        if not request.body:
            raise Exception

        current_filter = await self._get_gilter(request)
        aggregation = Aggregation(
            {
                "operation": request.body["aggregate"],
                "field": request.body.get("aggregate_field"),
                "groups": [{"field": request.body["group_by_field"]}],
            }
        )
        rows = await request.collection.aggregate(current_filter, aggregation)
        results: List[Dict[str, Union[str, int]]] = []
        for row in rows:
            key = row["group"][request.body["group_by_field"]]
            if ":" not in request.body["group_by_field"]:
                field = request.collection.get_field(request.body["group_by_field"])
                if cast(Column, field)["column_type"] == PrimitiveType.ENUM:
                    key = key.value
            results.append({"key": key, "value": row["value"]})
        return self._build_success_response(results)

    @check_method(RequestMethod.POST)
    @authenticate
    async def line(self, request: RequestCollection) -> Response:
        if not request.body:
            raise Exception

        current_filter = await self._get_gilter(request)
        aggregation = Aggregation(
            {
                "operation": request.body["aggregate"],
                "field": request.body.get("aggregate_field"),
                "groups": [{"field": request.body["group_by_date_field"], "operation": request.body["time_range"]}],
            }
        )
        rows = await request.collection.aggregate(current_filter, aggregation)
        values = {
            datetime.fromisoformat(row["group"][request.body["group_by_date_field"]]).date(): row["value"]
            for row in rows
            if row["group"][request.body["group_by_date_field"]] is not None
        }
        dates = list(values.keys())
        dates.sort()
        end = dates[-1]
        start = dates[0]
        datapoints: List[Dict[str, Union[date, Dict[str, int]]]] = []
        for dt in pd.date_range(  # type: ignore
            start=start, end=end, freq=self.FREQUENCIES[request.body["time_range"]].value
        ).to_pydatetime():
            datapoints.append(
                {
                    "label": dt.strftime(self.FORMAT[request.body["time_range"]]),
                    "values": {"value": values.get(dt.date(), 0)},
                }
            )
        return self._build_success_response(datapoints)

    @check_method(RequestMethod.POST)
    @authenticate
    async def leader_board(self, request: RequestCollection) -> Response:
        if not request.body:
            raise Exception
        aggregate = request.body["aggregate"]
        label_field = request.body["label_field"]
        relationship_field = request.body["relationship_field"]
        limit = request.body["limit"]
        aggregate_field = request.body.get("aggregate_field")
        if not aggregate_field:
            relation = SchemaUtils.get_to_many_relation(request.collection.schema, relationship_field)
            foreign_collection = relation["foreign_collection"]
            collection = request.collection.datasource.get_collection(foreign_collection)
            pks = SchemaUtils.get_primary_keys(collection.schema)
            aggregate_field = pks[0]
        current_filter = await self._get_gilter(request)
        aggregation = Aggregation(
            {
                "operation": aggregate,
                "field": f"{relationship_field}:{aggregate_field}",
                "groups": [{"field": label_field}],
            }
        )

        rows = await request.collection.aggregate(current_filter, aggregation, int(limit))
        results: List[Dict[str, Union[str, int]]] = []
        for row in rows:
            results.append({"key": row["group"][label_field], "value": row["value"]})
        return self._build_success_response(results)

    def _build_success_response(self, result: Any) -> Response:
        return build_success_response({"data": {"id": uuid1().hex, "type": "stats", "attributes": {"value": result}}})

    def _with_count_previous(self, tree: ConditionTree):
        use_interval_res: List[bool] = []

        def _use_interval_res(tree: ConditionTree) -> None:
            if isinstance(tree, ConditionTreeLeaf):
                use_interval_res.append(tree.use_interval_operator)

        tree.apply(_use_interval_res)
        return any(use_interval_res)

    async def _get_gilter(self, request: RequestCollection) -> Filter:
        scope_tree = await self.permission.get_scope(request, request.collection)
        return build_filter(request, scope_tree)

    async def _compute_value(self, request: RequestCollection, filter: Filter) -> int:
        if not request.body:
            raise Exception
        aggregation = Aggregation(
            {"operation": request.body["aggregate"], "field": request.body.get("aggregate_field")}
        )
        rows = await request.collection.aggregate(filter, aggregation)
        res = 0
        if len(rows):
            res = int(rows[0]["value"])
        return res



from typing import List, Optional

from forestadmin.agent_toolkit.utils.forest_schema.type import ServerValidationType, ValidationType
from forestadmin.datasource_toolkit.interfaces.fields import Operator, Validation


class FrontendValidationUtils:

    OPERATOR_VALIDATION_TYPE = {
        Operator.PRESENT: ValidationType.PRESENT,
        Operator.GREATER_THAN: ValidationType.GREATER_THAN,
        Operator.LESS_THAN: ValidationType.LESS_THAN,
        Operator.LONGER_THAN: ValidationType.LONGER_THAN,
        Operator.SHORTER_THAN: ValidationType.SHORTER_THAN,
        Operator.CONTAINS: ValidationType.CONTAINS,
        Operator.LIKE: ValidationType.LIKE,
    }

    @classmethod
    def convert_validation_list(cls, predicates: Optional[List[Validation]]):
        if not predicates:
            predicates = []

        res: List[ServerValidationType] = []
        for predicate in predicates:
            validation_type = cls.OPERATOR_VALIDATION_TYPE.get(predicate["operator"])
            if validation_type:
                res.append(
                    {
                        "type": validation_type.value,
                        "value": predicate.get("value"),
                        "message": None,
                    }
                )
        return res



import asyncio
import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from typing import Any, Awaitable, Dict, List, Tuple, cast

from forestadmin.agent_toolkit.resources.collections import BaseCollectionResource
from forestadmin.agent_toolkit.resources.collections.decorators import authenticate, authorize, check_method
from forestadmin.agent_toolkit.resources.collections.exceptions import CollectionResourceException
from forestadmin.agent_toolkit.resources.collections.filter import (
    FilterException,
    build_filter,
    build_paginated_filter,
    parse_condition_tree,
    parse_projection_with_pks,
    parse_selection_ids,
    parse_timezone,
)
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection, RequestCollectionException
from forestadmin.agent_toolkit.services.serializers.json_api import JsonApiException, JsonApiSerializer
from forestadmin.agent_toolkit.utils.context import (
    Request,
    RequestMethod,
    Response,
    build_client_error_response,
    build_no_content_response,
    build_success_response,
    build_unknown_response,
)
from forestadmin.agent_toolkit.utils.id import unpack_id
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import DatasourceException
from forestadmin.datasource_toolkit.interfaces.fields import (
    ManyToOne,
    OneToOne,
    Operator,
    is_column,
    is_many_to_many,
    is_many_to_one,
    is_one_to_many,
    is_one_to_one,
)
from forestadmin.datasource_toolkit.interfaces.query.aggregation import Aggregation
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.factory import ConditionTreeFactory
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.query.projections.factory import ProjectionFactory
from forestadmin.datasource_toolkit.interfaces.records import CompositeIdAlias, RecordsDataAlias
from forestadmin.datasource_toolkit.utils.collections import CollectionUtils
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils
from forestadmin.datasource_toolkit.validations.field import FieldValidatorException
from forestadmin.datasource_toolkit.validations.records import RecordValidator, RecordValidatorException
from typing_extensions import TypeGuard


def is_request_collection(request: Request) -> TypeGuard[RequestCollection]:
    return hasattr(request, "collection")


LiteralMethod = Literal["get", "list", "count", "add", "update", "delete", "delete_list", "update_list"]


class CrudResource(BaseCollectionResource):
    async def dispatch(self, request: Request, method_name: LiteralMethod) -> Response:
        method = getattr(self, method_name)
        try:
            request_collection = RequestCollection.from_request(request, self.datasource)
        except RequestCollectionException as e:
            return build_client_error_response([str(e)])
        return await method(request_collection)

    @check_method(RequestMethod.GET)
    @authenticate
    @authorize("read")
    async def get(self, request: RequestCollection) -> Response:
        collection = request.collection
        try:
            ids = unpack_id(collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        trees: List[ConditionTree] = [ConditionTreeFactory.match_ids(collection.schema, [ids])]
        scope_tree = await self.permission.get_scope(request)
        if scope_tree:
            trees.append(scope_tree)

        filter = PaginatedFilter({"condition_tree": ConditionTreeFactory.intersect(trees)})

        projections = ProjectionFactory.all(cast(Collection, collection))
        records = await collection.list(filter, projections)
        for name, schema in collection.schema["fields"].items():
            if is_many_to_many(schema) or is_one_to_many(schema):
                projections.append(f"{name}:id")  # type: ignore
                records[0][name] = None
        if not len(records):
            return build_unknown_response()
        else:
            schema = JsonApiSerializer.get(collection)
            try:
                dumped: Dict[str, Any] = schema(projections=projections).dump(records[0])  # type: ignore
            except JsonApiException as e:
                return build_client_error_response([str(e)])

        return build_success_response(dumped)

    @check_method(RequestMethod.POST)
    @authenticate
    @authorize("add")
    async def add(self, request: RequestCollection) -> Response:

        collection = request.collection
        schema = JsonApiSerializer.get(collection)
        try:
            data: RecordsDataAlias = schema().load(request.body)  # type: ignore
        except JsonApiException as e:
            return build_client_error_response([str(e)])

        record, one_to_one_relations = await self.extract_data(cast(Collection, collection), data)

        try:
            RecordValidator.validate(cast(Collection, collection), record)
        except RecordValidatorException as e:
            return build_client_error_response([str(e)])

        try:
            records = await collection.create([record])
        except DatasourceException as e:
            return build_client_error_response([str(e)])

        try:
            await self._link_one_to_one_relations(request, records[0], one_to_one_relations)
        except CollectionResourceException as e:
            return build_client_error_response([str(e)])

        return build_success_response(
            schema(projections=list(records[0].keys())).dump(records[0], many=False)  # type: ignore
        )

    @check_method(RequestMethod.GET)
    @authenticate
    @authorize("browse")
    async def list(self, request: RequestCollection) -> Response:
        scope_tree = await self.permission.get_scope(request)
        try:
            paginated_filter = build_paginated_filter(request, scope_tree)
        except FilterException as e:
            return build_client_error_response([str(e)])
        try:
            projections = parse_projection_with_pks(request)
        except DatasourceException as e:
            return build_client_error_response([str(e)])

        records = await request.collection.list(paginated_filter, projections)
        schema = JsonApiSerializer.get(request.collection)
        try:
            dumped: Dict[str, Any] = schema(projections=projections).dump(records, many=True)  # type: ignore
        except JsonApiException as e:
            return build_client_error_response([str(e)])
        return build_success_response(dumped)

    @check_method(RequestMethod.GET)
    @authenticate
    @authorize("browse")
    async def count(self, request: RequestCollection) -> Response:
        scope_tree = await self.permission.get_scope(request)
        filter = build_filter(request, scope_tree)
        aggregation = Aggregation({"operation": "Count"})
        result = await request.collection.aggregate(filter, aggregation)
        try:
            count = result[0]["value"]
        except IndexError:
            count = 0
        return build_success_response({"count": count})

    @check_method(RequestMethod.PUT)
    @authenticate
    @authorize("edit")
    async def update(self, request: RequestCollection) -> Response:
        collection = request.collection
        try:
            ids = unpack_id(collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        if request.body and "data" in request.body and "relationships" in request.body["data"]:
            del request.body["data"]["relationships"]

        schema = JsonApiSerializer.get(collection)
        try:
            data: RecordsDataAlias = schema().load(request.body)  # type: ignore
        except JsonApiException as e:
            return build_client_error_response([str(e)])
        try:
            RecordValidator.validate(cast(Collection, collection), data)
        except RecordValidatorException as e:
            return build_client_error_response([str(e)])

        trees: List[ConditionTree] = [ConditionTreeFactory.match_ids(collection.schema, [ids])]
        scope_tree = await self.permission.get_scope(request)
        if scope_tree:
            trees.append(scope_tree)

        filter = Filter({"condition_tree": ConditionTreeFactory.intersect(trees)})

        await collection.update(filter, data)
        projection = ProjectionFactory.all(cast(Collection, collection))
        records = await collection.list(PaginatedFilter.from_base_filter(filter), projection)

        schema = JsonApiSerializer.get(collection)
        try:
            dumped: Dict[str, Any] = schema(projections=projection).dump(records[0])  # type: ignore
        except JsonApiException as e:
            return build_client_error_response([str(e)])

        return build_success_response(dumped)

    @check_method(RequestMethod.DELETE)
    @authenticate
    @authorize("delete")
    async def delete(self, request: RequestCollection):
        collection = request.collection
        try:
            ids = unpack_id(collection.schema, request.pks)
        except (FieldValidatorException, CollectionResourceException) as e:
            return build_client_error_response([str(e)])

        await self._delete(request, [ids])
        return build_no_content_response()

    @check_method(RequestMethod.DELETE)
    @authenticate
    @authorize("delete")
    async def delete_list(self, request: RequestCollection):
        ids, exclude_ids = parse_selection_ids(request)
        await self._delete(request, ids, exclude_ids)
        return build_no_content_response()

    async def _delete(self, request: RequestCollection, ids: List[CompositeIdAlias], exclude_ids: bool = False):
        selected_ids = ConditionTreeFactory.match_ids(request.collection.schema, ids)
        if exclude_ids:
            selected_ids = selected_ids.inverse()
        trees = [selected_ids]
        query_param_condition_tree = parse_condition_tree(request)
        if query_param_condition_tree:
            trees.append(query_param_condition_tree)
        scope_tree = await self.permission.get_scope(request)
        if scope_tree:
            trees.append(scope_tree)

        await request.collection.delete(build_filter(request, ConditionTreeFactory.intersect(trees)))

    async def _link_one_to_one_relation(
        self, request: RequestCollection, record: RecordsDataAlias, relation: OneToOne, linked: RecordsDataAlias
    ):
        foreign_collection = self.datasource.get_collection(relation["foreign_collection"])
        scope = await self.permission.get_scope(request, foreign_collection)
        await self.permission.can(request, f"edit:{request.collection.name}")
        origin_value = record[relation["origin_key_target"]]

        # Break the old relation
        old_fk_owner = ConditionTreeLeaf(relation["origin_key"], Operator.EQUAL, origin_value)
        trees: List[ConditionTree] = [old_fk_owner]
        if scope:
            trees.append(scope)
        try:
            tz = parse_timezone(request)
        except FilterException as e:
            raise CollectionResourceException(str(e))

        await foreign_collection.update(
            Filter({"condition_tree": ConditionTreeFactory.intersect(trees), "timezone": tz}),
            {
                f'{relation["origin_key"]}': None,
            },
        )

        # Create the new relation
        new_fk_owner = ConditionTreeFactory.match_records(foreign_collection.schema, [linked])
        trees: List[ConditionTree] = [new_fk_owner]
        if scope:
            trees.append(scope)

        await foreign_collection.update(
            Filter({"condition_tree": ConditionTreeFactory.intersect(trees), "timezone": tz}),
            {
                f'{relation["origin_key"]}': origin_value,
            },
        )

    async def _link_one_to_one_relations(
        self, request: RequestCollection, record: RecordsDataAlias, relations: List[Tuple[OneToOne, RecordsDataAlias]]
    ):
        awaitables: List[Awaitable[None]] = []
        for relation, linked in relations:
            awaitables.append(self._link_one_to_one_relation(request, record, relation, linked))
        await asyncio.gather(*awaitables)

    async def extract_data(
        self, collection: Collection, data: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], List[Tuple[OneToOne, RecordsDataAlias]]]:
        record: Dict[str, Any] = {}
        one_to_one_relations: List[Tuple[OneToOne, RecordsDataAlias]] = []
        for field_name, value in data.items():
            field = collection.get_field(field_name)
            if is_column(field):
                record[field_name] = value
            elif is_many_to_one(field) or is_one_to_one(field):
                foreign_collection = self.datasource.get_collection(field["foreign_collection"])
                pk_names = SchemaUtils.get_primary_keys(foreign_collection.schema)
                if is_one_to_one(field):
                    one_to_one_relations.append((field, dict([(pk, value[i]) for i, pk in enumerate(pk_names)])))
                else:
                    field = cast(ManyToOne, field)
                    record[field["foreign_key"]] = await CollectionUtils.get_value(
                        cast(Collection, foreign_collection), [value], field["foreign_key_target"]
                    )

        return record, one_to_one_relations



import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import Any, Dict, List, Optional


class Data(TypedDict):
    type: str
    relationships: Dict[str, Any]
    attributes: Dict[str, Any]
    id: int
    links: Dict[str, Any]


class DumpedResult(TypedDict):
    data: List[Data]
    included: Dict[str, Any]
    meta: Optional[Dict[str, Any]]


def add_search_metadata(dumped: DumpedResult, search_value: str):
    if len(search_value.strip()) > 0:
        results_data = dumped["data"]
        decorators: Dict[int, Dict[str, Any]] = {}
        key = 0
        for result in results_data:
            search_fields: List[str] = []
            for field_name, value in result.items():
                if search_value.lower() in str(value).lower():
                    search_fields.append(field_name)
            if len(search_fields) > 0:
                decorators[key] = {"id": result["id"], "search": search_fields}
                key += 1
        dumped["meta"] = {"decorators": decorators}
    return dumped


"""
const decorators = resultsData.reduce((decorator, record: RecordData) => {
    const search = Object.keys(record.attributes).filter(attribute => {
        const value = record.attributes[attribute];

        return value && value.toString().toLowerCase().includes(searchValue.toLowerCase());
    });

    if (search.length === 0) {
        return decorator;
    }

    return { ...decorator, [Object.keys(decorator).length]: { id: record.id, search } };
    }, {});

    if (Object.values(decorators).length === 0) {
    return results;
    }
}

"""



from typing import Union

from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerSegment
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection


class SchemaSegmentGenerator:
    @staticmethod
    async def build(collection: Union[Collection, CustomizedCollection], name: str) -> ForestServerSegment:
        return {"id": f"{collection.name}.{name}", "name": name}



import json
import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from urllib.parse import urljoin

from forestadmin.agent_toolkit.options import Options
from forestadmin.agent_toolkit.resources.base import BaseResource
from forestadmin.agent_toolkit.resources.security.exceptions import AuthenticationException
from forestadmin.agent_toolkit.utils.authentication import ClientFactory, CustomClientOic
from forestadmin.agent_toolkit.utils.context import Request, Response
from forestadmin.agent_toolkit.utils.http import ForestHttpApi
from forestadmin.agent_toolkit.utils.token import build_jwt

LiteralMethod = Literal["authenticate", "callback"]


class Authentication(BaseResource):
    def __init__(self, options: Options):
        super(Authentication, self).__init__(options)
        self.callback_url = urljoin(options["agent_url"], "/forest/authentication/callback")

    async def dispatch(self, request: Request, method_name: LiteralMethod) -> Response:
        method = getattr(self, method_name)
        return await method(request)

    async def authenticate(self, request: Request) -> Response:
        client: CustomClientOic = await ClientFactory.build(self.callback_url, self.option)
        if not request.body:
            raise AuthenticationException("renderingId is missing in the request's body")
        try:
            rendering_id = int(request.body["renderingId"])
        except KeyError:
            raise AuthenticationException("renderingId is missing in the request's body")
        except ValueError:
            raise AuthenticationException("renderingId should be an integer")

        authorization_url = client.get_authorization_url(json.dumps({"renderingId": rendering_id}))

        return Response(
            status=200,
            body=json.dumps({"authorizationUrl": authorization_url}),
            headers={"content_type": "application/json"},
        )

    async def callback(self, request: Request) -> Response:
        client: CustomClientOic = await ClientFactory.build(self.callback_url, self.option)
        if not request.query:
            raise AuthenticationException("`state`should be sent to the callback endpoint")
        try:
            state = json.loads(request.query["state"])
        except KeyError:
            raise AuthenticationException("`state`should be sent to the callback endpoint")
        except ValueError:
            raise AuthenticationException("state should be a json")

        try:
            rendering_id = int(state["renderingId"])
        except KeyError:
            raise AuthenticationException("renderingId is missing in the callback state")
        except ValueError:
            raise AuthenticationException("renderingId should be an integer")

        authorization_response = client.get_parsed_response(request.query)
        access_token_response = client.get_access_token(authorization_response)
        rendering_authorization_response = await ForestHttpApi.get_rendering_authorization(
            rendering_id, access_token_response["access_token"], self.option
        )
        user = rendering_authorization_response["data"]["attributes"]

        token, body = build_jwt(
            {
                "id": rendering_authorization_response["data"]["id"],
                "email": user["email"],
                "first_name": user["first_name"],
                "last_name": user["last_name"],
                "team": user["teams"][0],
                "rendering_id": rendering_id,
            },
            self.option["auth_secret"],
        )

        return Response(
            status=200,
            body=json.dumps({"token": token, "tokenData": body}),
            headers={"content_type": "application/json"},
        )

    async def logout(self, request: Request) -> Response:
        return Response(status=204, body="", headers={})



from typing import Any, Dict, List, Optional, Union

from forestadmin.agent_toolkit.resources.actions.requests import ActionRequest, RequestActionException
from forestadmin.agent_toolkit.resources.collections import BaseCollectionResource
from forestadmin.agent_toolkit.resources.collections.crud import LiteralMethod
from forestadmin.agent_toolkit.resources.collections.decorators import authenticate, check_method
from forestadmin.agent_toolkit.resources.collections.filter import (
    build_filter,
    parse_condition_tree,
    parse_selection_ids,
)
from forestadmin.agent_toolkit.utils.context import (
    FileResponse,
    Request,
    RequestMethod,
    Response,
    build_client_error_response,
    build_json_response,
    build_success_response,
)
from forestadmin.agent_toolkit.utils.forest_schema.action_values import ForestValueConverter
from forestadmin.agent_toolkit.utils.forest_schema.generator_action import SchemaActionGenerator
from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerActionField
from forestadmin.datasource_toolkit.decorators.action.result_builder import ResultBuilder
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.factory import ConditionTreeFactory
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import Aggregator, ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter


class ActionResource(BaseCollectionResource):
    async def dispatch(self, request: Request, method_name: LiteralMethod) -> Response:
        method = getattr(self, method_name)
        try:
            request_collection = ActionRequest.from_request(request, self.datasource)
        except RequestActionException as e:
            return build_client_error_response([str(e)])
        return await method(request_collection)

    @check_method(RequestMethod.POST)
    @authenticate
    async def execute(self, request: ActionRequest) -> Union[FileResponse, Response]:
        if not request.body or not request.query:
            raise Exception
        filter = await self._get_records_selection(request)
        raw_data: Dict[str, Any] = request.body.get("data", {}).get("attributes", {}).get("values", {})  # type: ignore
        unsafe_data = ForestValueConverter.make_form_unsafe_data(raw_data)
        fields = await request.collection.get_form(request.action_name, raw_data, filter)
        data = ForestValueConverter.make_form_data(request.collection.datasource, unsafe_data, fields)
        result = await request.collection.execute(request.action_name, data, filter)
        if result["type"] == ResultBuilder.ERROR:
            return build_json_response(400, {"error": result["message"]})
        elif result["type"] == ResultBuilder.SUCCESS:
            key = "success"
            if result["format"] == "html":
                key = "html"
            return build_success_response({key: result["message"]})
        elif result["type"] == ResultBuilder.WEBHOOK:
            return build_success_response(
                {
                    "webhook": {
                        "url": result["url"],
                        "method": result["method"],
                        "headers": result["headers"],
                        "body": result["body"],
                    }
                }
            )
        elif result["type"] == ResultBuilder.FILE:
            return FileResponse(result["stream"], result["name"], result["mimeType"])

        elif result["type"] == ResultBuilder.REDIRECT:
            return build_success_response({"redirectTo": result["path"]})
        raise

    @check_method(RequestMethod.POST)
    @authenticate
    async def hook(self, request: ActionRequest) -> Response:
        if not request.body:
            raise Exception
        forest_fields: List[ForestServerActionField] = (
            request.body.get("data", {})  # type: ignore
            .get("attributes", {})  # type: ignore
            .get("fields", [])  # type: ignore
        )
        data: Optional[Dict[str, Any]] = None
        if forest_fields:
            data = ForestValueConverter.make_form_data_from_fields(request.collection.datasource, forest_fields)

        filter = await self._get_records_selection(request)
        fields = await request.collection.get_form(request.action_name, data, filter)

        return build_success_response(
            {
                "fields": [
                    await SchemaActionGenerator.build_field_schema(request.collection.datasource, field)
                    for field in fields
                ]
            }
        )

    async def _get_records_selection(self, request: ActionRequest) -> Filter:
        if not request.body:
            raise Exception()
        trees: List[ConditionTree] = []
        selection_ids, exclude_ids = parse_selection_ids(request)
        selected_ids = ConditionTreeFactory.match_ids(request.collection.schema, selection_ids)
        if exclude_ids:
            selected_ids = selected_ids.inverse()

        if selected_ids != ConditionTreeBranch(Aggregator.OR, []):
            trees.append(selected_ids)
        query_param_condition_tree = parse_condition_tree(request)
        if query_param_condition_tree:
            trees.append(query_param_condition_tree)
        scope_tree = await self.permission.get_scope(request)
        if scope_tree:
            trees.append(scope_tree)

        if trees:
            condition_tree = ConditionTreeFactory.intersect(trees)
        else:
            condition_tree = None
        filter = build_filter(request, condition_tree)
        attributes: Dict[str, Any] = request.body.get("data", {}).get("attributes", {})
        if "parent_association_name" in attributes:
            pass
        return filter



from typing import Any, Awaitable, Callable, TypeVar, Union

from forestadmin.agent_toolkit.resources.collections import BaseCollectionResource
from forestadmin.agent_toolkit.resources.collections.requests import RequestCollection
from forestadmin.agent_toolkit.services.permissions import PermissionServiceException
from forestadmin.agent_toolkit.utils.context import (
    FileResponse,
    Request,
    RequestMethod,
    Response,
    User,
    build_method_not_allowed_response,
)
from jose import JWTError, jwt

BoundRequest = TypeVar("BoundRequest", bound=Request)
BoundResource = TypeVar("BoundResource", bound=BaseCollectionResource)
BoundRequestCollection = TypeVar("BoundRequestCollection", bound=RequestCollection)


def authenticate(fn: Callable[["BoundResource", BoundRequestCollection], Awaitable[Union[FileResponse, Response]]]):
    async def wrapped2(self: "BoundResource", request: BoundRequestCollection) -> Union[FileResponse, Response]:
        if not request.headers:
            return Response(status=401)

        try:
            authorization = request.headers["Authorization"]
        except KeyError:
            return Response(status=401)

        try:
            token = authorization.split("Bearer ")[1]
        except IndexError:
            return Response(status=401)

        try:
            user = jwt.decode(token, self.option["auth_secret"])
        except JWTError:
            return Response(status=401)

        request.user = User(
            rendering_id=int(user["rendering_id"]),
            user_id=int(user["id"]),
            tags=user.get("tags", {}),
            email=user["email"],
            first_name=user["first_name"],
            last_name=user["last_name"],
            team=user["team"],
        )
        return await fn(self, request)

    return wrapped2


def authorize(action: str):
    def wrapper(fn: Callable[["BoundResource", BoundRequestCollection], Awaitable[Any]]):
        async def wrapped1(self: "BoundResource", request: BoundRequestCollection) -> Union[FileResponse, Response]:
            try:
                await self.permission.can(request, f"{action}:{request.collection.name}")
            except PermissionServiceException as e:
                return Response(status=e.STATUS, body=e.message)

            return await fn(self, request)

        return wrapped1

    return wrapper


def check_method(method: RequestMethod):
    def wrapper(fn: Callable[["BoundResource", BoundRequestCollection], Awaitable[Union[FileResponse, Response]]]):
        async def wrapped(self: "BoundResource", request: BoundRequestCollection) -> Union[FileResponse, Response]:
            if request.method != method:
                return build_method_not_allowed_response()
            return await fn(self, request)

        return wrapped

    return wrapper



import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class RoleOptions(TypedDict):
    forest_server_url: str
    env_secret: str
    is_production: bool
    permission_cache_duration: int



import enum
import sys

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

from typing import Any, List, Optional, Union

from forestadmin.datasource_toolkit.interfaces.fields import ColumnAlias
from typing_extensions import NotRequired


class ValidationType(enum.Enum):
    PRESENT = "is present"
    GREATER_THAN = "is greater than"
    LESS_THAN = "is less than"
    BEFORE = "is before"
    AFTER = "is after"
    LONGER_THAN = "is longer than"
    SHORTER_THAN = "is shorter than"
    CONTAINS = "contains"
    LIKE = "is like"


class ServerValidationType(TypedDict):
    type: ValidationType
    value: Optional[Any]
    message: Optional[Any]


LiteralHasOne = Literal["HasOne"]
LiteralHasMany = Literal["HasMany"]
LiteralBelongsTo = Literal["BelongsTo"]
LiteralBelongsToMany = Literal["BelongsToMany"]

RelationServer = Union[LiteralHasOne, LiteralHasMany, LiteralBelongsTo, LiteralBelongsToMany]


class ForestServerField(TypedDict, total=False):
    field: str
    type: ColumnAlias
    defaultValue: Any
    enums: Optional[List[str]]
    integration: None
    isFilterable: bool
    isPrimaryKey: bool
    isReadOnly: bool
    isRequired: bool
    isSortable: bool
    isVirtual: bool
    reference: Optional[str]
    inverseOf: Optional[str]
    relationship: RelationServer
    validations: List[ServerValidationType]


LiteralPage = Literal["page"]


class ForestServerCollection(TypedDict):
    name: str
    icon: None
    integration: None
    isReadOnly: bool
    isSearchable: bool
    isVirtual: bool
    onlyForRelationships: bool
    paginationType: LiteralPage
    actions: NotRequired[List[Any]]
    fields: List[ForestServerField]
    segments: NotRequired[List[Any]]


class ForestServerActionHooks(TypedDict):
    load: bool
    change: List[Any]


class ForestServerActionField(TypedDict):
    value: Any
    defaultValue: Any
    description: Optional[str]
    enums: Optional[List[str]]
    field: str
    hook: Optional[str]
    isReadOnly: bool
    isRequired: bool
    reference: Optional[str]
    type: Union[ColumnAlias, Literal["File"]]
    widget: Optional[Literal["belongsto select", "file picker"]]


class ForestServerAction(TypedDict):
    id: str
    name: str
    type: Literal["single", "bulk", "global"]
    baseUrl: Optional[str]
    endpoint: str
    httpMethod: Literal["POST"]
    redirect: Any
    download: bool
    fields: List[ForestServerActionField]
    hooks: ForestServerActionHooks


"""
export type ForestServerSegment = {
  id: string;
  name: string;
};
"""


class ForestServerSegment(TypedDict):
    id: str
    name: str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forestadmin',
 'forestadmin.agent_toolkit',
 'forestadmin.agent_toolkit.resources',
 'forestadmin.agent_toolkit.resources.actions',
 'forestadmin.agent_toolkit.resources.collections',
 'forestadmin.agent_toolkit.resources.security',
 'forestadmin.agent_toolkit.services',
 'forestadmin.agent_toolkit.services.permissions',
 'forestadmin.agent_toolkit.services.serializers',
 'forestadmin.agent_toolkit.utils',
 'forestadmin.agent_toolkit.utils.forest_schema']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8,<4.0',
 'cachetools>=5.2,<6.0',
 'forestadmin-datasource-toolkit>=0.1.0-beta.14,<0.2.0',
 'marshmallow-jsonapi>=0.24.0,<1.0',
 'oic>=1.4,<2.0',
 'python-jose>=3.3,<4.0',
 'typing-extensions>=4.2.0,<5.0',
 'tzdata>=2022.6,<2022.7']

extras_require = \
{':python_full_version < "3.7.1"': ['pandas==1.1.5'],
 ':python_full_version >= "3.7.1" and python_version < "3.8"': ['pandas==1.3.5'],
 ':python_version < "3.9"': ['backports.zoneinfo[tzdata]>=0.2.1,<0.3.0'],
 ':python_version >= "3.8"': ['pandas>=1.4.2,<1.5.0']}

setup_kwargs = {
    'name': 'forestadmin-agent-toolkit',
    'version': '0.1.0b14',
    'description': '',
    'long_description': '',
    'author': 'Valentin Monté',
    'author_email': 'valentinm@forestadmin.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List, Union, cast

from forestadmin.agent_toolkit.utils.forest_schema.action_values import ForestValueConverter
from forestadmin.agent_toolkit.utils.forest_schema.generator_field import SchemaFieldGenerator
from forestadmin.agent_toolkit.utils.forest_schema.type import ForestServerAction, ForestServerActionField
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.decorators.collections import CustomizedCollection
from forestadmin.datasource_toolkit.interfaces.actions import Action, ActionField, ActionFieldType
from forestadmin.datasource_toolkit.interfaces.fields import Column, PrimitiveType
from forestadmin.datasource_toolkit.utils.schema import SchemaUtils


class SchemaActionGenerator:

    DUMMY_FIELDS = [
        ForestServerActionField(
            field="Loading...",
            type=SchemaFieldGenerator.build_column_type(PrimitiveType.STRING),
            isReadOnly=True,
            defaultValue="Form is loading",
            value=None,
            description="",
            enums=None,
            hook=None,
            isRequired=False,
            reference=None,
            widget=None,
        )
    ]

    @classmethod
    async def build(
        cls, prefix: str, collection: Union[Collection, CustomizedCollection], name: str
    ) -> ForestServerAction:
        schema = collection.schema["actions"][name]
        idx = list(collection.schema["actions"].keys()).index(name)
        slug = name.lower().replace(r"[^a-z0-9-]+", "-")
        return ForestServerAction(
            id=f"{collection.name}-{idx}-{slug}",
            type=schema.scope.value.lower(),
            name=name,
            baseUrl=None,
            endpoint=f"/{prefix}/_actions/{collection.name}/{idx}/{slug}",
            httpMethod="POST",
            redirect=None,
            download=bool(schema.generate_file),
            fields=await cls.build_fields(collection, schema, name),
            hooks={"load": not schema.static_form, "change": ["changeHook"]},
        )

    @classmethod
    async def build_field_schema(
        cls, datasource: Datasource[Collection], field: ActionField
    ) -> ForestServerActionField:
        output: ForestServerActionField = {
            "value": ForestValueConverter.value_to_forest(field, field["value"]),
            "defaultValue": None,
            "description": field["description"],
            "enums": None,
            "field": field["label"],
            "hook": None,
            "isReadOnly": field["is_read_only"] or False,
            "isRequired": field["is_required"] or True,
            "reference": None,
            "type": PrimitiveType.STRING,
            "widget": None,
        }
        if field["type"] == ActionFieldType.COLLECTION:
            collection: Collection = datasource.get_collection(field["collection_name"])  # type: ignore
            pk = SchemaUtils.get_primary_keys(collection.schema)[0]
            pk_schema = cast(Column, collection.get_field(pk))
            output["type"] = SchemaFieldGenerator.build_column_type(pk_schema["column_type"])  # type: ignore
            output["reference"] = f"{collection.name}.{pk}"
        elif field["type"] == ActionFieldType.ENUM_LIST:
            output["type"] = PrimitiveType.ENUM
        elif field["type"] == ActionFieldType.NUMBER_LIST:
            output["type"] = PrimitiveType.NUMBER
        elif field["type"] == ActionFieldType.FILE:
            output["type"] = "File"
        else:
            output["type"] = PrimitiveType(field["type"].value)

        if field["type"] in [ActionFieldType.ENUM, ActionFieldType.ENUM_LIST]:
            output["enums"] = field["enum_values"]
        if not isinstance(output["type"], str):
            output["type"] = SchemaFieldGenerator.build_column_type(output["type"])

        if field["watch_changes"]:
            output["hook"] = "changeHook"

        return ForestServerActionField(**output)

    @classmethod
    async def build_fields(
        cls, collection: Union[Collection, CustomizedCollection], action: Action, name: str
    ) -> List[ForestServerActionField]:
        if not action.static_form:
            return cls.DUMMY_FIELDS
        fields = await collection.get_form(name, None, None)
        new_fields: List[ForestServerActionField] = []
        for field in fields:
            new_fields.append(await cls.build_field_schema(collection.datasource, field))
        return new_fields



import sys

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict


class AgentStackMeta(TypedDict, total=False):
    engine: Literal["python"]
    engine_version: str
    database_type: str
    orm_version: str


class AgentMeta(TypedDict):
    liana: str
    liana_version: str
    stack: AgentStackMeta


class Options(TypedDict):
    application_url: str
    agent_url: str
    auth_secret: str
    prefix: str
    env_secret: str
    forest_server_url: str
    is_production: bool
